package de.monarchy.guideme.data;

import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Map;
import java.util.Set;

import android.content.Context;
import android.content.SharedPreferences;
import android.hardware.Sensor;
import android.hardware.SensorManager;
import android.location.Criteria;
import android.location.Location;
import android.location.LocationManager;
import android.os.Handler;
import android.os.Looper;
import android.os.Message;
import android.preference.PreferenceManager;
import android.util.Log;
import android.widget.Toast;
import de.monarchy.guideme.R;
import de.monarchy.guideme.pois.POIProvider;
import de.monarchy.guideme.routing.RoutingComponent;
import de.monarchy.guideme.routing.RoutingListener;
import de.monarchy.guideme.routing.nutiteq.NutiteqRouting;
import de.monarchy.guideme.util.GeoCoordinate;
import de.monarchy.guideme.util.POI;
import de.monarchy.guideme.util.RoutePoint;

/**
 * PersonalLocation is the main service of the program, providing information
 * about position and orientation of the mobile device. After binding to this
 * class, any other component can be notified about changes in position,
 * orientation and routing via handlers. PersonalLocation therefore provides
 * two lists. MovementHandlers are getting updates about orientation and
 * position changes, respectively indicated by the constants POSITION_CHANGED
 * and ORIENTATION_CHANGED as described below.
 * RoutingHandlers receive information when routing calls succeed or fail.
 * Those, too, are indicated by constants ROUTING_FAILED and ROUTING_DONE. 
 * 
 * POSITION_CHANGED will provide a GeoCoordinate as m.obj.
 * ORIENTATION_CHANGED provides a float[] as m.obj as described in SensorEvent.
 * ROUTING_DONE delivers a RoutingPoint[] as m.obj.
 * ROUTING_FAILED is only a notice and does not carry any additional data.
 * 
 * Every given data should not be manipulated directly. If a Handler needs to do
 * so, it should build a deep copy of the object to avoid side effects.
 * 
 * @author dstuerze
 * 
 */
public class PersonalLocation implements Runnable, RoutingListener {

  /** minimal time span for updating positioning information */
  protected static final int MIN_TIME = 1000;
  /** minimal distance for updating positioning information */
  protected static final int MIN_DISTANCE = 5;
  /** minimal change in angles for receiving orientation changes */
  protected static final float MIN_ANGLE = 4;

  /** kills the thread */
  public static final int KILL_COMMAND = -100;
  /** status telling that the orientation of the device changed */
  public static final int ORIENTATION_CHANGED = 1;
  /** status telling that the position of the device changed */
  public static final int POSITION_CHANGED = 2;
  /** status telling that a new route was calculated */
  public static final int ROUTING_DONE = 4;
  /** status telling that a routing call was unsuccessful */
  public static final int ROUTING_FAILED = 5;
  /** status asking for routing service from actual position */
  public static final int CALCULATE_ROUTE_TO = 32;
  /** status asking for routing service from a given point */
  public static final int CALCULATE_ROUTE_FROM_TO = 33;

  /** variable expressing the user's preferred language for pois */
  public static String preferred_language = "de";

  // file for saving data
  private static final String SAVE_FILE = "guideme_data";

  // list of registered handlers for movements
  private Set<Handler> movementHandlerList = new HashSet<Handler>();

  /** context, needed for service connections */
  public Context context;

  /** handler for message queue */
  public Handler handler;

  // internal class for receiving movement information
  private PositionListener listener;

  // local information about orientation and position
  private float[] orientation = { 0f, 0f, 0f };
  private GeoCoordinate position = new GeoCoordinate(0,0);

  // list of registered handlers for routing
  private static Set<Handler> routingHandlerList = new HashSet<Handler>();

  // routing service
  private static RoutingComponent routes;

  // route information
  private RoutePoint[] route;

  // registered providers for POI reception
  private Map<POIProvider, GeoCoordinate> poiProviders = new HashMap<POIProvider, GeoCoordinate>();
  private Handler poiDelegate = new Handler(new POIHandler());

  // poi data
  private Set<POI> pois = new HashSet<POI>();
  private int distance = 5000;

  /**
   * Creates a new data provider.
   * 
   * @param c
   *          context the object is living in
   */
  public PersonalLocation(Context c) {
    context = c;
    Log.v("GuideMe", "PersonalLocation created");

    // find POI providers
    String[] providerStrings = c.getResources().getStringArray(
        R.array.poiProviders);
    for (String pr : providerStrings) {
      try {
        // create them
        POIProvider p = (POIProvider) Class.forName(pr).newInstance();
        poiProviders.put(p, new GeoCoordinate(0., 0.));

      } catch (Exception e) {
        e.printStackTrace();
      }
    }

    // register routing service
    if (routes == null)
      routes = new NutiteqRouting();

    // load initial pois
    if (position != null)
      updatePOIs();
  }

  protected void finalize() {
  }

  /**
   * Returns the orientation of the device in an float array. This array is
   * similar to the one returned by SensorEvents, so for additional information
   * you should look up there.
   * 
   * @return orientation of the device
   */
  public float[] getOrientation() {
    return orientation;
  }

  /**
   * Sets the orientation of the device. This method is only to be calles by a
   * suitable listener. It is thread safe, so multiple calls to this method will
   * be processed in a correct order.
   * 
   * @param orientation
   *          new orientation as described in SensorEvent
   */
  public synchronized void setOrientation(float[] orientation) {
    boolean change = false;
    if (Math.abs(orientation[0] - this.orientation[0]) > MIN_ANGLE) {
      this.orientation[0] = orientation[0];
      change = true;
    }
    if (Math.abs(orientation[1] - this.orientation[1]) > MIN_ANGLE) {
      this.orientation[1] = orientation[1];
      change = true;
    }

    if (Math.abs(orientation[2] - this.orientation[2]) > MIN_ANGLE) {
      this.orientation[2] = orientation[2];
      change = true;
    }
    if (change)
      notifyHandlers(ORIENTATION_CHANGED);

  }

  /**
   * Returns the last route calculated by the service.
   * 
   * @return last known route
   */
  public RoutePoint[] getRoute() {
    return route;
  }

  /**
   * Removes the calculated route. Will not inform listeners.
   */
  public void clearRoute() {
    route = null;
  }

  /**
   * Returns the current position of the device as an GeoCoordinate.
   * 
   * @return position
   */
  public GeoCoordinate getPosition() {
    return position;
  }

  /**
   * Sets the position of the device. This method is only to be calles by a
   * suitable listener. It is thread safe, so multiple calls to this method will
   * be processed in a correct order.
   * 
   * @param position
   *          new position of the device
   */
  public synchronized void setPosition(GeoCoordinate position) {
    Log.d("GuideMe", "PersonalLocation new location=" + position.toString());
    if (this.position == null) {
      this.position = new GeoCoordinate(position);
    } else if (this.position.distanceTo(position) * GeoCoordinate.EARTH_RADIUS > 1.) {
      this.position = new GeoCoordinate(position);
      updatePOIs();
      notifyHandlers(POSITION_CHANGED);
    }
  }

  /**
   * Calculates a route between the given points. The method just calls the
   * registered routing service and lies still while waiting for a result.
   * 
   * @param from
   *          starting point
   * @param to
   *          destination point
   */
  public void calculateRoute(GeoCoordinate from, GeoCoordinate to) {
    Log.d("GuideMe", "PersonalLocation routeTo=" + to.toString());
    routes.calculateRoute(this, from, to);
  }

  /**
   * Convenience method for finding a route between the actual position and the
   * given destination point. Similar call as calculateRoute(getPosition(), to).
   * 
   * @param to
   *          destination point.
   */
  public void calculateRoute(GeoCoordinate to) {
    calculateRoute(position, to);
  }

  /**
   * Adds a Handler to this service, so external objects can retain information
   * about the position of the device. The method is thread safe, so multiple
   * calls will not corrupt the list of handlers. New handlers go in last
   * position of the handler list. Every handler can only be added once to the
   * service.
   * 
   * @param m
   *          new handler
   * @return true if the handler could be added
   */
  public synchronized boolean addMovementHandler(Handler m) {
    return movementHandlerList.add(m);
  }

  /**
   * Removes a handler from this service.
   * 
   * @param m
   *          handler to be removed
   * @return true if removal was successful
   */
  public synchronized boolean removeMovementHandler(Handler m) {
    return movementHandlerList.remove(m);
  }

  /**
   * Adds a Handler to this service, so external objects can retain information
   * about changing routing information. The method is thread safe, so multiple
   * calls will not corrupt the list of handlers. New handlers go in last
   * position of the handler list. Every handler can only be added once to the
   * service.
   * 
   * @param m
   *          new handler
   * @return true if the handler could be added
   */
  public synchronized boolean addRoutingHandler(Handler m) {
    return routingHandlerList.add(m);
  }

  /**
   * Removes a handler from this service.
   * 
   * @param m
   *          handler to be removed
   * @return true if removal was successful
   */
  public synchronized boolean removeRoutingHandler(Handler m) {
    return routingHandlerList.remove(m);
  }

  /**
   * Private method to send notifications to signed handlers. Information about
   * the messages sent can be found in the documentation of the setters in this
   * class.
   * 
   * @param state
   *          type of change
   */
  private void notifyHandlers(final int state) {
    Message m = null;
    Set<Handler> internal;
    synchronized (this) {
      switch (state) {
      // in case of movement
      case ORIENTATION_CHANGED:
      case POSITION_CHANGED:
        internal = movementHandlerList;
        break;
      // in case of routing
      case ROUTING_DONE:
      case ROUTING_FAILED:
        internal = routingHandlerList;
        break;
      // in case of wrong value
      default:
        internal = new HashSet<Handler>();
      }
    }
    // iterate handlers
    for (Handler h : internal) {
      // lazily create one message
      if (m == null) {
        m = h.obtainMessage();
        switch (state) {
        // notify about orientation
        case ORIENTATION_CHANGED:
          m.obj = orientation;
          break;
        // notify about position
        case POSITION_CHANGED:
          m.obj = position;
          break;
        case ROUTING_DONE:
          m.obj = route;
          break;
        case ROUTING_FAILED:
          m.obj = null;
          break;
        }
        m.what = state;
        m.sendToTarget();
      }
    }
  }

  /**
   * Registers the service's listener to the best location service available. To
   * avoid multiple calls, the listener will first be removed from all location
   * services and then added to the one matching the following criteria the
   * best: - high accuracy - no additional costs - low power consumption
   */
  public void registerPositionListener() {
    if (listener == null)
      return;
    // find location manager
    LocationManager lm = (LocationManager) context
        .getSystemService(Context.LOCATION_SERVICE);
    lm.removeUpdates(listener);
    Log.i("GuideMe", "PersonalLocation position updates removed");
    String prov = null;
    if (lm.isProviderEnabled(LocationManager.GPS_PROVIDER)) {
      // try to obtain gps data
      prov = LocationManager.GPS_PROVIDER;
      lm.requestLocationUpdates(prov, MIN_TIME, MIN_DISTANCE, listener);
    } else {
      // if gps is unavailable, get best location criteria possible
      Criteria crit = new Criteria();
      crit.setAccuracy(Criteria.ACCURACY_FINE);
      crit.setCostAllowed(false);
      crit.setPowerRequirement(Criteria.POWER_LOW);
      // find provider
      prov = lm.getBestProvider(crit, true);
      // register listener
      lm.requestLocationUpdates(prov, MIN_TIME, MIN_DISTANCE, listener);
    }

    Location l = lm.getLastKnownLocation(prov);
    if (l != null)
      setPosition(new GeoCoordinate(l.getLatitude(), l.getLongitude()));
  }

  @Override
  public void onError(int code) {
    Log.i("GuideMe", "PersonalLocation routing failed");
    notifyHandlers(ROUTING_FAILED);
  }

  @Override
  public void onRouteCalculated(RoutePoint[] points) {
    Log.d("GuideMe", "PersonalLocation routing succeeded");
    route = points;
    notifyHandlers(ROUTING_DONE);
  }

  /**
   * This method is responsible to call all registered POIProviders and get an
   * update.
   */
  private void updatePOIs() {
    // go through all providers
    Set<Map.Entry<POIProvider, GeoCoordinate>> entries = poiProviders
        .entrySet();
    for (Map.Entry<POIProvider, GeoCoordinate> entry : entries) {
      GeoCoordinate ort = entry.getValue();
      final POIProvider prov = entry.getKey();
      // if moved more than 1000 metres, start a search
      if (ort.distanceTo(position) * GeoCoordinate.EARTH_RADIUS > 1000.) {
        // as we do not know how long this may take, do it in a new thread
        Runnable r = new Runnable() {
          public void run() {
            // receive data
            POI[] erg = prov.receivePOIs(position, distance, null);
            // give it back to the system
            Message m = new Message();
            m.obj = erg;
            poiDelegate.sendMessage(m);
          }
        };
        new Thread(r).start();
        entry.setValue(new GeoCoordinate(position));
      }
    }
  }

  /**
   * Returns an array of all known POIs.
   * 
   * @return pois in an array
   */
  public POI[] getPOIs() {
    return pois.toArray(new POI[0]);
  }

  /**
   * Internal handler implementation. Used to put new pois where they belong.
   * 
   * @author dstuerze
   * 
   */
  private class POIHandler implements Handler.Callback {
    public boolean handleMessage(Message m) {
      POI[] poiData = (POI[]) m.obj;
      for (int i = 0; i < poiData.length; i++) {
        pois.add(poiData[i]);
      }
      PersonalLocation.this.notifyHandlers(POSITION_CHANGED);
      return true;
    }
  }

  public void run() {
    Looper.prepare();

    handler = new Handler(Looper.myLooper()) {
      public void handleMessage(Message m) {
        switch (m.what) {
        case POSITION_CHANGED:
          setPosition((GeoCoordinate) m.obj);
          break;
        case ORIENTATION_CHANGED:
          setOrientation((float[]) m.obj);
          break;
        case CALCULATE_ROUTE_FROM_TO:
          GeoCoordinate from = (GeoCoordinate) m.getData().getSerializable(
              "from");
          GeoCoordinate to = (GeoCoordinate) m.getData().getSerializable("to");
          calculateRoute(from, to);
          break;
        case CALCULATE_ROUTE_TO:
          calculateRoute((GeoCoordinate) m.obj);
          break;
        case KILL_COMMAND:
          kill();
          break;
        }
      }
    };

    // add position listener
    listener = new PositionListener(this);
    registerPositionListener();
    onStart();

    Looper.loop();
    Log.v("GuideMe", "PersonalLocation end run");
  }

  /**
   * Cuts the location service off the system services, so that it will no
   * longer get informations about position or orientation updates.
   */
  public void onStop() {
    Log.v("GuideMe", "PersonalLocation stopped");
    // location updates
    LocationManager lm = (LocationManager) context
        .getSystemService(Context.LOCATION_SERVICE);
    lm.removeUpdates(listener);
    // orientation updates
    SensorManager sm = (SensorManager) context
        .getSystemService(Context.SENSOR_SERVICE);
    sm.unregisterListener(listener);
    // saving
    SharedPreferences p = PreferenceManager
        .getDefaultSharedPreferences(context);
    boolean save = p.getBoolean(context.getResources().getString(
        R.string.pref_key_pers), false);
    if (save)
      save();
  }

  /**
   * Registers the position listener at system services so the thread is able to
   * maintain information.
   */
  public void onStart() {
    Log.v("GuideMe", "PersonalLocation started");
    // location updates
    if (listener != null)
      registerPositionListener();
    // orientation updates
    SensorManager sm = (SensorManager) context
        .getSystemService(Context.SENSOR_SERVICE);
    sm.registerListener(listener, sm.getDefaultSensor(Sensor.TYPE_ORIENTATION),
        SensorManager.SENSOR_DELAY_UI);
    // get preferred language
    SharedPreferences p = PreferenceManager
        .getDefaultSharedPreferences(context);
    preferred_language = p.getString(context.getResources().getString(
        R.string.pref_key_lang), "de");
    // loading
    boolean load = p.getBoolean(context.getResources().getString(
        R.string.pref_key_pers), false);
    if (load)
      load();

  }

  /**
   * Saves the data on hard drive.
   */
  private void save() {
    try {
      FileOutputStream fo = null;
      fo = context.openFileOutput(SAVE_FILE, Context.MODE_PRIVATE);
      ObjectOutputStream out = new ObjectOutputStream(fo);
      Log.d("GuideMe", "PersonalLocation save");
      out.writeObject(position);
      out.writeObject(orientation);
      out.writeObject(route);
      out.writeObject(pois);
      out.flush();
      out.close();
    } catch (IOException e) {
      Log.i("GuideMe", "PersonalLocation save failed");
      e.printStackTrace();
      Toast t = Toast.makeText(context, context.getResources().getString(
          R.string.pl_save_fail), Toast.LENGTH_SHORT);
      t.show();
    }
  }

  /**
   * Loads an old instance from hard drive.
   */
  @SuppressWarnings("unchecked")
  private void load() {
    try {
      FileInputStream fi = null;
      fi = context.openFileInput(SAVE_FILE);
      ObjectInputStream in = new ObjectInputStream(fi);
      Log.d("GuideMe", "PersonalLocation load");
      position = (GeoCoordinate) in.readObject();
      Log.v("GuideMe", "PersonalLocation load position=" + position);
      orientation = (float[]) in.readObject();
      Log.v("GuideMe", "PersonalLocation load orientation=" + orientation);
      route = (RoutePoint[]) in.readObject();
      Log.v("GuideMe", "PersonalLocation load route=" + route);
      pois = (Set<POI>) in.readObject();
      Log.v("GuideMe", "PersonalLocation load pois=" + pois);
      in.close();
      notifyHandlers(POSITION_CHANGED);
      notifyHandlers(ORIENTATION_CHANGED);
      if (route != null)
        notifyHandlers(ROUTING_DONE);

    } catch (Exception e) {
      Log.i("GuideMe", "PersonalLocation load failed");
      e.printStackTrace();
      Toast t = Toast.makeText(context, context.getResources().getString(
          R.string.pl_load_fail), Toast.LENGTH_SHORT);
      t.show();
    }
  }

  /**
   * Kills the thread, saves the data first.
   */
  private void kill() {
    Log.v("GuideMe", "PersonalLocation deleted");
    onStop();
    movementHandlerList.clear();
    routingHandlerList.clear();
    Looper.myLooper().quit();

  }

}
