package de.monarchy.guideme.main;

import java.net.URL;
import java.util.HashSet;
import java.util.Set;

import android.app.Activity;
import android.content.Context;
import android.content.Intent;
import android.content.SharedPreferences;
import android.content.SharedPreferences.Editor;
import android.graphics.PorterDuff;
import android.net.Uri;
import android.os.Bundle;
import android.os.Handler;
import android.os.Message;
import android.os.PowerManager;
import android.preference.PreferenceManager;
import android.util.Log;
import android.view.Menu;
import android.view.MenuInflater;
import android.view.MenuItem;
import android.view.View;
import android.view.Window;
import android.view.View.OnClickListener;
import android.view.ViewGroup.LayoutParams;
import android.widget.ImageView;
import android.widget.RelativeLayout;
import de.monarchy.guideme.R;
import de.monarchy.guideme.data.PersonalLocation;
import de.monarchy.guideme.util.GeoCoordinate;
import de.monarchy.guideme.util.POI;
import de.monarchy.guideme.util.RoutePoint;
import de.monarchy.guideme.util.WebPOI;
import de.monarchy.guideme.view.CombinationView;

/**
 * GuidMe's main activity. Responsible for showing all necessary views,
 * preferences, menu items etc. Also, this class informs those views via
 * callbacks about changes in orientation, positioning, routing or viewing.
 * 
 * @author dstuerze
 * 
 */
public class MainActivity extends Activity {

  /** constant value for routing activity start */
  protected static final int ROUTING_REQUEST = 0x1000;

  // data controllers
  private InformationHandler sc;
  private PersonalLocation persloc;

  // views
  private CombinationView cv;
  private ImageView map, cam;

  // preference related
  private PowerManager.WakeLock wl;
  private SharedPreferences preferences;

  public void onCreate(Bundle savedInstanceState) {
    super.onCreate(savedInstanceState);

    // fill preferences
    PreferenceManager.setDefaultValues(this, R.xml.preferences, false);
    preferences = PreferenceManager.getDefaultSharedPreferences(this);

    // frameless window
    requestWindowFeature(Window.FEATURE_NO_TITLE);

    // build the layout
    RelativeLayout rl = new RelativeLayout(this);

    // main view with camera and map
    cv = new CombinationView(this);
    RelativeLayout.LayoutParams cvp = new RelativeLayout.LayoutParams(
        LayoutParams.FILL_PARENT, LayoutParams.FILL_PARENT);
    rl.addView(cv, cvp);

    // button for switching camera on and off
    cam = new ImageView(this);
    cam.setImageResource(R.drawable.camera);
    // at start on or off?
    boolean camOn = preferences.getBoolean(getResources().getString(
        R.string.pref_key_cam), true);
    // add to layout in lower left corner
    RelativeLayout.LayoutParams camp = new RelativeLayout.LayoutParams(
        LayoutParams.WRAP_CONTENT, LayoutParams.WRAP_CONTENT);
    camp.addRule(RelativeLayout.ALIGN_PARENT_BOTTOM);
    camp.addRule(RelativeLayout.ALIGN_PARENT_LEFT);
    rl.addView(cam, camp);
    // on clicking camera mode will change
    cam.setOnClickListener(new ActivationListener(cam, camOn) {
      public void onClick(View v) {
        super.onClick(v);
        // save new state to preferences
        Editor e = preferences.edit();
        e.putBoolean(getResources().getString(R.string.pref_key_cam), super.on);
        e.commit();
        cv.setCamActivated(this.on);
      }
    });

    // button for switching map on/off
    map = new ImageView(this);
    map.setImageResource(R.drawable.world_globe);
    // at start on or off?
    boolean mapOn = preferences.getBoolean(getResources().getString(
        R.string.pref_key_map), true);
    // add to layout in lower right corner
    RelativeLayout.LayoutParams mapp = new RelativeLayout.LayoutParams(
        LayoutParams.WRAP_CONTENT, LayoutParams.WRAP_CONTENT);
    mapp.addRule(RelativeLayout.ALIGN_PARENT_BOTTOM);
    mapp.addRule(RelativeLayout.ALIGN_PARENT_RIGHT);
    rl.addView(map, mapp);
    // on clicking map mode will change
    map.setOnClickListener(new ActivationListener(map, mapOn) {
      public void onClick(View v) {
        super.onClick(v);
        // save new state to preferences
        Editor e = preferences.edit();
        e.putBoolean(getResources().getString(R.string.pref_key_map), super.on);
        e.commit();
        cv.setMapActivated(super.on);
      }
    });

    // confirm layout
    setContentView(rl);

    // location service start
    persloc = new PersonalLocation(this);
    new Thread(persloc).start();
    Thread.yield();

    // data service handler
    sc = new InformationHandler(persloc);
    Log.d("GuideMe", "MainActivity onCreate");

  }

  /**
   * This listener is responsible for changing an image view's tint, indicating
   * it is activated or not.
   * 
   * @author dstuerze
   * 
   */
  private class ActivationListener implements OnClickListener {
    // data
    protected boolean on = true;
    private ImageView source;

    /**
     * The listener is bound to the given view, with the given state as default.
     * 
     * @param src
     *          ImageView to be changed
     * @param activated
     *          initial state
     */
    ActivationListener(ImageView src, boolean activated) {
      source = src;
      on = activated;
    }

    @Override
    public void onClick(View v) {
      source.clearColorFilter();
      if (on)
        // switching off
        source.setColorFilter(v.getResources().getColor(R.color.off_tint),
            PorterDuff.Mode.valueOf("SRC_ATOP"));
      else
        // switching on
        source.setColorFilter(v.getResources().getColor(R.color.on_tint),
            PorterDuff.Mode.valueOf("SRC_ATOP"));
      on = !on;
      source.invalidate();
    }

  }

  public void onRestart() {
    Log.d("GuideMe", "MainActivity onRestart");
    super.onRestart();
  }

  public void onResume() {
    Log.d("GuideMe", "MainActivity onResume");
    super.onResume();
    // start and initialize the postioning service
    persloc.onStart();
    sc.addOrientationChangeListener(cv);
    sc.addPositionChangeListener(cv);
    sc.addRoutingListener(cv);
    sc.refreshData();
    cv.onStart();

    // map on or off?
    boolean mapOn = preferences.getBoolean(getResources().getString(
        R.string.pref_key_map), true);
    if (mapOn)
      map.setColorFilter(getResources().getColor(R.color.on_tint),
          PorterDuff.Mode.valueOf("SRC_ATOP"));
    else
      map.setColorFilter(getResources().getColor(R.color.off_tint),
          PorterDuff.Mode.valueOf("SRC_ATOP"));

    // cam on or off?
    boolean camOn = preferences.getBoolean(getResources().getString(
        R.string.pref_key_cam), true);
    if (camOn)
      cam.setColorFilter(getResources().getColor(R.color.on_tint),
          PorterDuff.Mode.valueOf("SRC_ATOP"));
    else
      cam.setColorFilter(getResources().getColor(R.color.off_tint),
          PorterDuff.Mode.valueOf("SRC_ATOP"));

    // inform the view about what is activated
    cv.setMapActivated(mapOn);
    cv.setCamActivated(camOn);

    // preference telling if the screen should not be locked
    PowerManager pm = (PowerManager) getSystemService(Context.POWER_SERVICE);
    wl = pm.newWakeLock(PowerManager.SCREEN_BRIGHT_WAKE_LOCK, "My Tag");
    if (preferences.getBoolean(getResources().getString(
        R.string.pref_key_screen), true))
      wl.acquire();
  }

  public void onPause() {
    super.onPause();
    // let the screen be dark
    if (wl.isHeld())
      wl.release();
    // stop position updates
    persloc.onStop();
    sc.removeOrientationChangeListener(cv);
    sc.removePositionChangeListener(cv);
    sc.removeRoutingListener(cv);
    // clear the views
    cv.onStop();
    Log.d("GuideMe", "MainActivity onPause");
  }

  public void onStop() {
    super.onStop();
    Log.d("GuideMe", "MainActivity onStop");
  }

  public void onDestroy() {
    super.onDestroy();
    Log.d("GuideMe", "MainActivity destroyed");
    Message.obtain(persloc.handler, PersonalLocation.KILL_COMMAND).sendToTarget();
    cv.onStop();
  }

  public boolean onCreateOptionsMenu(Menu menu) {
    // create the main menu
    super.onCreateOptionsMenu(menu);
    MenuInflater inf = new MenuInflater(getApplication());
    inf.inflate(R.menu.main_menu, menu);
    return true;
  }

  public boolean onOptionsItemSelected(MenuItem item) {
    switch (item.getItemId()) {
    case R.id.start_routing:
      // start routing activity to search a new destination
      Intent routeSearch = new Intent(this, RoutingActivity.class);
      startActivityForResult(routeSearch, ROUTING_REQUEST);
      break;
    case R.id.restart_routing:
      // redo the routing from this position
      sc.findRoute(sc.route[sc.route.length - 1]);
      break;
    case R.id.stop_routing:
      // delete the route
      sc.route = null;
      cv.removeRoute();
      persloc.clearRoute();
      break;
    case R.id.preferences:
      // show preferences
      Intent preferenceScreen = new Intent(this, PreferenceActivity.class);
      startActivity(preferenceScreen);
      break;
    default:
      Log.i("GuideMe", "MainMenu: no such element");
    }
    return false;
  }

  public boolean onContextItemSelected(MenuItem item) {
    // if clicked on an item in context menu
    POI info = (POI) item.getMenuInfo();
    switch (item.getItemId()) {
    case R.id.item_web:
      // for webpois, open the web page
      if (info instanceof WebPOI) {
        URL url = ((WebPOI) info).getUrl();
        Intent i = new Intent(Intent.ACTION_VIEW, Uri.parse(url.toString()));
        startActivity(i);
      }
      break;
    case R.id.item_route:
      // use as route destination
      sc.findRoute(info.getLocation());
      break;
    default:
    }

    return super.onContextItemSelected(item);
  }

  public void onActivityResult(int requestCode, int resultCode, Intent data) {
    super.onActivityResult(requestCode, resultCode, data);
    if (requestCode == ROUTING_REQUEST) {
      // this should be the only case: we started the routing activity
      switch (resultCode) {
      case Activity.RESULT_OK:
        // if a destination was chosen, get a route
        Bundle content = data.getExtras();
        GeoCoordinate target = (GeoCoordinate) content.get("target");
        sc.findRoute(target);
        break;
      case Activity.RESULT_CANCELED:
        // else don't
        break;
      default:
        Log.i("GuideMe", "ROUTING_REQUEST: illegal result");
      }
    }
  }

  /**
   * This class is a bridge between the activity and the location service. It
   * provides data for everyone interested in that. Communication with the
   * service is via messages.
   * 
   * @author dstuerze
   * 
   */
  class InformationHandler extends Handler {

    // data provider
    protected PersonalLocation locator;

    // observers
    protected Set<OrientationChangeListener> orientationListenerList;
    protected Set<PositionChangeListener> positionListenerList;
    protected Set<RoutingListener> routingListenerList;

    // actual data
    protected RoutePoint[] route;
    protected Set<POI> pois;
    protected GeoCoordinate position = new GeoCoordinate(0,0);
    protected float[] orientation;

    /**
     * Initializes a new InformationHandler for the given data provider.
     * 
     * @param locator
     *          data source
     */
    public InformationHandler(PersonalLocation locator) {
      // base data
      this.locator = locator;
      orientationListenerList = new HashSet<OrientationChangeListener>();
      positionListenerList = new HashSet<PositionChangeListener>();
      routingListenerList = new HashSet<RoutingListener>();
      pois = new HashSet<POI>();

      // register for updates
      locator.addMovementHandler(this);
      locator.addRoutingHandler(this);

      // get initial data
      position = locator.getPosition();
      orientation = locator.getOrientation();
      route = locator.getRoute();
      POI[] poiArray = locator.getPOIs();
      for (int i = 0; i < poiArray.length; i++) {
        pois.add(poiArray[i]);
      }

      // inform the listeners
      refreshData();
      Log.d("GuideMe", "InformationHandler initialized");
    }

    public void handleMessage(Message m) {
      if (m.what == PersonalLocation.ORIENTATION_CHANGED) {
        // if the orientation of the device changed...
        orientation = (float[]) m.obj;
        callOrientationListeners();
      }

      if (m.what == PersonalLocation.POSITION_CHANGED) {
        // if the position of the device changed...
        if (m.obj != null)
          position = (GeoCoordinate) m.obj;
        POI[] pois = locator.getPOIs();
        for (POI p : pois)
          this.pois.add(p);
        callPositionListeners();
      }
      if (m.what == PersonalLocation.ROUTING_FAILED
          || m.what == PersonalLocation.ROUTING_DONE) {
        // if routing has taken place...
        route = (RoutePoint[]) m.obj;
        callRoutingListeners();
      }

    }

    /**
     * Convenience method for informing all listeners.
     */
    public void refreshData() {
      Log.d("GuideMe", "InformationHandler: refresh");
      if (position != null) {
        callPositionListeners();
        Log.v("GuideMe", "InformationHandler: refresh position");
      }
      if (orientation != null) {
        callOrientationListeners();
        Log.v("GuideMe", "InformationHandler: refresh orientation");
      }
      callRoutingListeners();
      Log.v("GuideMe", "InformationHandler: refresh routing");
    }

    /**
     * Calls every orientation listener and informs them about the device's
     * orientation.
     */
    private void callOrientationListeners() {
      OrientationChangeListener[] listener;
      synchronized (this) {
        listener = orientationListenerList
            .toArray(new OrientationChangeListener[0]);
      }
      for (OrientationChangeListener o : listener) {
        o.stateChanged(orientation);
      }
    }

    /**
     * Calls every routing listener and informs them about the actual route.
     */
    private void callRoutingListeners() {
      RoutingListener[] listener;
      synchronized (this) {
        listener = routingListenerList.toArray(new RoutingListener[0]);
      }
      for (RoutingListener o : listener) {
        o.routingDone(route);
      }
    }

    /**
     * Calls every position listener and informs them about the device's
     * position.
     */
    private void callPositionListeners() {
      PositionChangeListener[] listener;
      POI[] pois;
      synchronized (this) {
        listener = positionListenerList.toArray(new PositionChangeListener[0]);
        pois = this.pois.toArray(new POI[0]);
      }
      for (PositionChangeListener o : listener) {
        o.stateChanged(position, pois);
      }

    }

    /**
     * Calaculates a route to the given position. The job itself will be done by
     * the service.
     * 
     * @param g
     *          position the route will go to
     */
    public void findRoute(GeoCoordinate g) {
      locator.handler.obtainMessage(PersonalLocation.CALCULATE_ROUTE_TO, g)
          .sendToTarget();
    }

    /**
     * Adds a new OrientationChangeListener.
     * 
     * @param o
     *          new listener
     */
    public void addOrientationChangeListener(OrientationChangeListener o) {
      orientationListenerList.add(o);
    }

    /**
     * Removes an OrientationChangeListener.
     * 
     * @param o
     *          listener to be removed
     */
    public void removeOrientationChangeListener(OrientationChangeListener o) {
      orientationListenerList.remove(o);
    }

    /**
     * Adds a new PositionChangeListener.
     * 
     * @param o
     *          new listener
     */
    public void addPositionChangeListener(PositionChangeListener o) {
      positionListenerList.add(o);
    }

    /**
     * Removes an PositionChangeListener.
     * 
     * @param o
     *          listener to be removed
     */
    public void removePositionChangeListener(PositionChangeListener o) {
      positionListenerList.remove(o);
    }

    /**
     * Adds a new RoutingListener.
     * 
     * @param o
     *          new listener
     */
    public void addRoutingListener(RoutingListener o) {
      routingListenerList.add(o);
    }

    /**
     * Removes a RoutingListener.
     * 
     * @param o
     *          listener to be removed
     */
    public void removeRoutingListener(RoutingListener o) {
      routingListenerList.remove(o);
    }

  }

}
