package de.monarchy.guideme.main;

import android.app.Activity;
import android.app.AlertDialog;
import android.content.DialogInterface;
import android.content.Intent;
import android.content.DialogInterface.OnClickListener;
import android.os.Bundle;
import android.util.Log;
import android.view.View;
import android.view.Window;
import android.view.ViewGroup.LayoutParams;
import android.widget.EditText;
import android.widget.ImageButton;
import android.widget.LinearLayout;
import android.widget.RelativeLayout;
import de.monarchy.guideme.R;
import de.monarchy.guideme.map.MapWrapper;
import de.monarchy.guideme.map.TouchDelegate;
import de.monarchy.guideme.map.nutiteq.NutiteqWrapper;
import de.monarchy.guideme.routing.GeoCodingListener;
import de.monarchy.guideme.routing.android.AndroidGeoCoding;
import de.monarchy.guideme.routing.osm.OSMGeoCoding;
import de.monarchy.guideme.util.GeoCoordinate;
import de.monarchy.guideme.util.PointOfInterest;

/**
 * RoutingActivity is responsible for giving the user an opportunity to define a
 * new target for routing. This can either be done by clicking on a map or by
 * searching for a location.
 * 
 * @author dstuerze
 * 
 */
public class RoutingActivity extends Activity implements GeoCodingListener {

  /** distance, up to which possible destinations are found. */
  private static final int MAX_DIST = 20000;

  // mapping device
  private MapWrapper map;
  // view
  private LinearLayout layout;
  private View mapView;
  // map touching delegate
  private TouchDelegate delegate;

  // which target was selected?
  private int selectedItem;

  @Override
  public void onCreate(Bundle savedInstanceState) {
    super.onCreate(savedInstanceState);
    // main layouting
    requestWindowFeature(Window.FEATURE_NO_TITLE);
    layout = new LinearLayout(this);
    layout.setOrientation(LinearLayout.VERTICAL);
    RelativeLayout container = new RelativeLayout(this);

    // needed for the map
    map = NutiteqWrapper.getInstance(this);
    mapView = map.getView();
    delegate = new ClickDelegate();

    // search fields, activating button
    View.inflate(this, R.layout.search, container);
    LinearLayout.LayoutParams p = new LinearLayout.LayoutParams(
        LayoutParams.FILL_PARENT, LayoutParams.WRAP_CONTENT);
    layout.addView(container, 0, p);
    setContentView(layout);

    ImageButton im = (ImageButton) findViewById(R.id.search_button);
    im.setOnClickListener(new View.OnClickListener() {
      @Override
      public void onClick(View v) {
        EditText e = (EditText) RoutingActivity.this
            .findViewById(R.id.search_text);
        String search = e.getText().toString();
        // search for addresses
        PointOfInterest[] pois = new AndroidGeoCoding(RoutingActivity.this)
            .findNear(search, map.getCenter(), MAX_DIST);
        // search for places
        PointOfInterest[] pois2 = new OSMGeoCoding().findNear(search, map
            .getCenter(), MAX_DIST);
        // let user choose
        onPlacesFound(search, concat(pois, pois2));
      }
    });
  }

  public void onResume() {
    super.onResume();
    Log.d("GuideMe", "RoutingActivity onResume");
    // set the map up for scrolling, zooming and reacting
    map.setScrollable(true);
    map.positionZoomControls(RelativeLayout.CENTER_HORIZONTAL,
        RelativeLayout.ALIGN_PARENT_BOTTOM);
    map.showViewDirection(false);
    map.setTouchDelegate(delegate);
    mapView.setVisibility(View.VISIBLE);

    // add the view below the search fields
    LinearLayout.LayoutParams par = new LinearLayout.LayoutParams(
        LayoutParams.FILL_PARENT, LayoutParams.FILL_PARENT);
    layout.addView(mapView, 1, par);
    layout.setVisibility(View.VISIBLE);
    layout.invalidate();
  }

  public void onPause() {
    super.onPause();
    layout.setVisibility(View.INVISIBLE);
    // free the map view, as it is a singleton
    layout.removeView(mapView);
    map.clearTouchDelegate();
  }

  @Override
  protected void onDestroy() {
    Log.d("GuideMe", "RoutingActivity onDestroy");
    // clean up
    super.onDestroy();
    if (mapView != null) {
      mapView = null;
    }
    if (map != null) {
      map = null;
    }
  }

  @Override
  public void onPlacesFound(String search, final PointOfInterest... places) {
    if (places.length == 0) {
      // if our search didn't deliver any places, just tell you're sorry
      String title = getResources().getString(R.string.geo_choose_title);
      String message = getResources().getString(R.string.geo_choose_message);
      String button = getResources().getString(R.string.geo_choose_info);
      // in a dialog
      AlertDialog dialog = new AlertDialog.Builder(this).setMessage(message)
          .setTitle(title).create();
      dialog.setButton(DialogInterface.BUTTON_NEUTRAL, button, new OnClickListener() {
        @Override
        public void onClick(DialogInterface dialog, int which) {
          dialog.dismiss();
        }
      });
      dialog.show();
      return;
    }
    // if there are results, show their string representations
    String[] strings = new String[places.length];
    Log.v("GuideMe", "RoutingActivity places: " + places.length);
    for (int i = 0; i < places.length; i++) {
      strings[i] = places[i].getDescription();
      Log.v("GuideMe", "RoutingActivity place[" + i + "]: " + strings[i]);
    }
    // the user will have to choose one of them
    AlertDialog dialog = new AlertDialog.Builder(this).setSingleChoiceItems(
        strings, -1, new OnClickListener() {

          @Override
          public void onClick(DialogInterface dialog, int which) {
            // remember selection
            selectedItem = which;
          }
        }).create();
    // set title and other textes
    String atext = getResources().getString(R.string.geo_choose_title);
    dialog.setTitle(atext);
    String btext = getResources().getString(R.string.geo_choose_positive);
    dialog.setButton(DialogInterface.BUTTON_POSITIVE, btext, new OnClickListener() {
      @Override
      public void onClick(DialogInterface dialog, int which) {
        Intent answer = new Intent();
        if (selectedItem < 0 || selectedItem >= places.length) {
          // do not find route
          RoutingActivity.this.setResult(Activity.RESULT_CANCELED, answer);
        } else {
          // find a route
          GeoCoordinate arg0 = places[selectedItem];
          answer.putExtra("target", arg0);
          RoutingActivity.this.setResult(Activity.RESULT_OK, answer);
        }
        dialog.dismiss();
        finish();
      }
    });

    btext = getResources().getString(R.string.geo_choose_negative);
    dialog.setButton(DialogInterface.BUTTON_NEGATIVE, btext, new OnClickListener() {
      @Override
      public void onClick(DialogInterface dialog, int which) {
        // user decided not to have found a fitting location
        dialog.dismiss();
      }
    });
    dialog.show();
  }

  @Override
  public void onError(int code) {
  }

  /**
   * Concatenates multiple arrays of PointOfInterest. Does not make a deep copy!
   * 
   * @param interests
   *          arrays
   * @return concatenated array
   */
  private PointOfInterest[] concat(PointOfInterest[]... interests) {

    // count the total number of pois
    int numbers = 0;
    for (int i = 0; i < interests.length; i++)
      numbers += interests[i].length;

    // copy them in one small array
    PointOfInterest[] result = new PointOfInterest[numbers];
    numbers = 0;
    for (int i = 0; i < interests.length; i++) {
      for (int j = 0; j < interests[i].length; j++) {
        result[numbers] = interests[i][j];
        numbers++;
      }
    }
    return result;
  }

  /**
   * A delegation class to react to touch events on the map-
   * 
   * @author dstuerze
   * 
   */
  private class ClickDelegate implements TouchDelegate {

    @Override
    public void onTouch(GeoCoordinate arg0) {
      // the clicked position is where the user wants to go
      Intent answer = new Intent();
      answer.putExtra("target", arg0);
      RoutingActivity.this.setResult(Activity.RESULT_OK, answer);
      finish();
    }
  }

}