package de.monarchy.guideme.map;

import android.view.View;
import de.monarchy.guideme.util.GeoCoordinate;
import de.monarchy.guideme.util.POI;
import de.monarchy.guideme.util.RoutePoint;

/**
 * MapWrapper: A class to implement an api wrapping mapping component. This will
 * be used as a controller for an underlying map view.
 * 
 * @author dstuerze
 * 
 */
public abstract class MapWrapper {

  // map view attributes about activity, size, scrolling, viewing and touching
  protected boolean isDrawing;
  protected boolean isFullScreen;
  protected boolean isScrollable;
  protected boolean showViewDirection;
  protected TouchDelegate touchDelegate;

  /**
   * The mapping base class MapWrapper is initialized with all internal
   * attributes set to true.
   */
  public MapWrapper() {
    isDrawing = true;
    isScrollable = true;
    showViewDirection = true;
    isFullScreen = true;
  }

  /**
   * Callback method for when the device's orientation has changed. The
   * orientation will be given in an array with three values as described by
   * Android's SensorEvent.
   * 
   * @param arg
   *          new orientation
   */
  public abstract void onOrientationChanged(float[] arg);

  /**
   * Callback method indicating a change of the device's orientation happened.
   * With the new position, an array of POIs in a range around this location is
   * submitted.
   * 
   * @param arg
   *          new position
   * @param arg1
   *          POIs around that spot
   */
  public abstract void onPositionChanged(GeoCoordinate arg, POI[] arg1);

  /**
   * Callback method indicating a new route has been calculated. If a routing
   * request failed, this method's parameter will be indicating it by being
   * null.
   * 
   * @param route
   *          array with route instructions
   */
  public abstract void setRoute(RoutePoint[] route);

  /**
   * Convenience method for removing the route to be mapped.
   */
  public abstract void removeRoute();

  /**
   * Returns the map view the controller has power over.
   * @return map view
   */
  public abstract View getView();

  /**
   * Sets zoom controls of the map view to the particular values. The parameters
   * shall be constants from RelativeLayout.LayoutParams indicating the position
   * in respect to the parent (which is the map view itself).
   * @param x horizontal position
   * @param y vertical position
   */
  public abstract void positionZoomControls(int x, int y);

  /**
   * Sets whether the map has to be drawn.
   * @param b true, if the map view should be drawn.
   */
  public void setDrawing(boolean b) {
    isDrawing = b;
  }

  /**
   * Sets, whether the map should give a visual indicator where the device is
   * pointing to.
   * @param b true if advice is wanted
   */
  public void showViewDirection(boolean b) {
    showViewDirection = b;
  }

  /**
   * Allows the controller to inform the view about being in fullscreen mode.
   * @param b true if map is filling the whole parent.
   */
  public void setFullScreen(boolean b) {
    isFullScreen = b;
  }

  /**
   * Indicates whether the map view should have the ability to scroll.
   * @param b true if scrolling is allowed
   */
  public void setScrollable(boolean b) {
    isScrollable = b;
  }

  /**
   * Returns whether the related view is scrollable.
   * @return true if scrollable, false else
   */
  public boolean isScrollable() {
    return isScrollable;
  }

  /**
   * Sets the touch delegate for the map.
   * @param d new touch delegate reacting to touches on the view
   */
  public void setTouchDelegate(TouchDelegate d) {
    touchDelegate = d;
  }

  /**
   * Removes the touch delegate.
   */
  public void clearTouchDelegate() {
    touchDelegate = null;
  }

  /**
   * Centers the map view at the given coordinate.
   * @param g position of the new center
   */
  public abstract void centerAt(GeoCoordinate g);

  /**
   * Returns the center position of the map.
   * @return momentary center position of the view in world coordinates
   */
  public abstract GeoCoordinate getCenter();

}
