package de.monarchy.guideme.pois.wikipedia;

import java.net.URL;

import android.content.Context;
import android.graphics.drawable.Drawable;
import android.view.ContextMenu;
import android.view.MenuInflater;
import android.view.View;
import android.widget.ImageView;
import de.monarchy.guideme.R;
import de.monarchy.guideme.util.GeoCoordinate;
import de.monarchy.guideme.util.PointOfInterest;
import de.monarchy.guideme.util.WebPOI;

/**
 * WikipediaArticle is a special class to maintain information about points of
 * interest gathered from wikipedia.
 * 
 * @author dstuerze
 * 
 */
public class WikipediaArticle extends PointOfInterest implements WebPOI {

  /** */
  private static final long serialVersionUID = 4497000926790723250L;

  /** article's URL */
  private final URL url;

  /** a short part of content */
  private String content;

  private transient ImageView view;

  private static transient Drawable image;

  /**
   * Creates a POI for wikipedia articles. As additional information, an url
   * referring to the article has to be added.
   * 
   * @param lat
   *          latitude of the POI
   * @param lon
   *          longitude of the POI
   * @param descr
   *          name of the article
   * @param url
   *          article url
   */
  public WikipediaArticle(double lat, double lon, String descr, URL url) {
    super(lat, lon, descr);
    this.url = url;
  }

  /**
   * Creates a POI for wikipedia articles. As additional information, an url
   * referring to the article has to be added.
   * 
   * @param gc
   *          coordinate of the POI
   * @param descr
   *          name of the article
   * @param url
   *          article url
   */
  public WikipediaArticle(GeoCoordinate gc, String descr, URL url) {
    super(gc, descr);
    this.url = url;
  }

  /**
   * Creates a POI for wikipedia articles. As additional information, an url
   * referring to the article and a short piece of content should be added.
   * 
   * @param lat
   *          latitude of the POI
   * @param lon
   *          longitude of the POI
   * @param descr
   *          name of the article
   * @param url
   *          article url
   * @param text
   *          short additional information about the POI
   */
  public WikipediaArticle(double lat, double lon, String descr, URL url,
      String text) {
    super(lat, lon, descr);
    this.url = url;
    content = text;
  }

  /**
   * Creates a POI for wikipedia articles. As additional information, an url
   * referring to the article and a short piece of content should be added.
   * 
   * @param gc
   *          coordinate of the POI
   * @param descr
   *          name of the article
   * @param url
   *          article url
   * @param text
   *          short additional information about the POI
   */
  public WikipediaArticle(GeoCoordinate gc, String descr, URL url, String text) {
    super(gc, descr);
    this.url = url;
    content = text;
  }

  /**
   * Sets the content of this article to the given String.
   * 
   * @param text
   *          new content
   */
  public void setContent(String text) {
    content = text;
  }

  /**
   * Returns the content of this article. As there is no error handling
   * involved, possible return values include empty strings and null.
   * 
   * @return content of the article if set, null else
   */
  public String getContent() {
    return content;
  }

  /**
   * Returns the full URL of this article.
   * 
   * @return url
   */
  public URL getUrl() {
    return url;
  }

  public boolean equals(Object o) {
    if (o == null)
      return false;
    if (o.getClass() != this.getClass())
      return false;
    WikipediaArticle w = (WikipediaArticle) o;
    return getDescription().equals(w.getDescription());

  }

  public int hashCode() {
    return getDescription().hashCode();
  }

  public View getView(Context c) {
    // create the view's icon exactly once for every view
    if (image == null) {
      image = c.getResources().getDrawable(R.drawable.wiki_icon);
    }
    if (view == null) {
      view = new ImageView(c) {
        protected void onCreateContextMenu(ContextMenu menu) {
          // add a context menu
          MenuInflater m = new MenuInflater(getContext());
          m.inflate(R.menu.poi_context_menu, menu);
          menu.setHeaderTitle(R.string.poi_context_menu_title);
        }

        protected ContextMenu.ContextMenuInfo getContextMenuInfo() {
          return WikipediaArticle.this;
        }
      };
      view.setImageDrawable(image);
    }
    return view;
  }

}
