package de.monarchy.guideme.pois.wikipedia;

import java.io.IOException;
import java.net.MalformedURLException;
import java.net.URL;

import javax.xml.parsers.ParserConfigurationException;
import javax.xml.parsers.SAXParser;
import javax.xml.parsers.SAXParserFactory;

import org.xml.sax.InputSource;
import org.xml.sax.SAXException;
import org.xml.sax.XMLReader;

import android.util.Log;
import de.monarchy.guideme.data.PersonalLocation;
import de.monarchy.guideme.pois.POIProvider;
import de.monarchy.guideme.util.GeoCoordinate;
import de.monarchy.guideme.util.POI;

/**
 * WikipediaProvider is a POIProvider implementation for wikipedia articles. It
 * uses a webservice by geonames.com for its purposes. As the webservice states,
 * wikipedia articles in over 240 languages can be obtained. for that, the user
 * has to choose one particular language at instantiation time.
 * 
 * @author dstuerze
 * 
 */
public class WikipediaProvider implements POIProvider {

  @Override
  public POI[] receivePOIs(GeoCoordinate where, int radius, String what) {
    // getting a SAX xml parser and reader
    SAXParserFactory factory = SAXParserFactory.newInstance();
    XMLReader xml;
    try {
      SAXParser parser = factory.newSAXParser();
      xml = parser.getXMLReader();
    } catch (SAXException e) {
      // as all following: if something goes wrong, just
      // return an empty list
      return new POI[0];
    } catch (ParserConfigurationException e) {
      return new POI[0];
    }
    // content handler implementation
    WikiHandler x = new WikiHandler();
    xml.setContentHandler(x);
    // webservice url
    URL url = parseUrl(where, radius);
    if (url == null)
      return new POI[0];
    InputSource input = new InputSource();
    try {
      // open the stream and parse
      Log.w("WikiProvider", "call: " + url.toString());
      input.setByteStream(url.openStream());
      xml.parse(input);
      return x.getEntries();
    } catch (IOException e) {
      return new POI[0];
    } catch (SAXException e) {
      return new POI[0];
    }
  }

  /**
   * Creates a webservice url for a specific location and radius.
   * 
   * @param point
   *          central position
   * @param radius
   *          radius around the position (in metres)
   * @return webservice url or null, if an error occured
   */
  private URL parseUrl(GeoCoordinate point, int radius) {
    StringBuffer url = new StringBuffer();
    url.append("http://ws.geonames.org/findNearbyWikipedia");
    // latitude
    String lat = String.valueOf(point.getLatitude());
    url.append("?lat=");
    url.append(lat);
    // longitude
    String lon = String.valueOf(point.getLongitude());
    url.append("&lng=");
    url.append(lon);
    // radius
    radius /= 1000;
    url.append("&radius=");
    url.append(Double.toString(radius));
    // language (if none, english is taken)
    if (PersonalLocation.preferred_language != null) {
      url.append("&lang=");
      url.append(PersonalLocation.preferred_language);
    }
    // maximum number of articles - not too many
    url.append("&maxRows=100");
    try {
      URL wUrl = new URL(url.toString());
      return wUrl;
    } catch (MalformedURLException e) {
      return null;
    }
  }

}
