package de.monarchy.guideme.routing;

import de.monarchy.guideme.util.GeoCoordinate;
import de.monarchy.guideme.util.PointOfInterest;

/**
 * Simple interface for using external geocoding services. A class implementing
 * this interface can either be providing data synchronously or 
 * asynchronously. It is not required to implement both ways, in that case, a 
 * UnsupportedOperationException is thrown. 
 * @author dstuerze
 *
 */
public interface GeoCodingComponent {
  /** searches for this address without bounds */
  public static final int EVERYWHERE = -1;

  /** Error message indicating the geocoding service is not responding */
  public static final int ERROR_SERVICE_DOWN = 1;

  /** Error message indicating a result could not be obtained */
  public static final int ERROR_NOT_FOUND = 2;

  /**
   * Searches places defined by a search string in an area around a certain
   * location. The method will call a geocoding service and deliver an array of
   * possible matches. The search is restricted to an area near a given
   * location. The maximal distance is denoted by the parameter maxDistance,
   * which can be set to the constant value EVERYWHERE to search without
   * restrictions.
   * 
   * @param search
   *          place that is searched
   * @param location
   *          location near the place
   * @param maxDistance
   *          maximum distance between location and place
   * @return array of possible matches
   * @throws UnsupportedOperationException
   *           if a synchronous call is not supported
   */
  public PointOfInterest[] findNear(String search, GeoCoordinate location,
      int maxDistance) throws UnsupportedOperationException;

  /**
   * Searches places defined by a search string in an area around a certain
   * location. The method will call a geocoding service and deliver an array of
   * possible matches. The search is restricted to an area near a given
   * location. The maximal distance is denoted by the parameter maxDistance,
   * which can be set to the constant value EVERYWHERE to search without
   * restrictions. After the call is completed, the given listener receives an
   * answer.
   * 
   * @param l
   *          RoutingListener the result is delivered to
   * @param search
   *          place that is searched
   * @param location
   *          location near the place
   * @param maxDistance
   *          maximum distance between location and place
   * @throws UnsupportedOperationException
   *           if an asynchronous call is not supported
   */
  public void findNear(GeoCodingListener l, String search,
      GeoCoordinate location, int maxDistance)
      throws UnsupportedOperationException;
}
