package de.monarchy.guideme.routing;

import de.monarchy.guideme.util.GeoCoordinate;
import de.monarchy.guideme.util.RoutePoint;

/**
 * Simple interface for using external routing services. A class implementing
 * this interface can either be providing routes synchronously or
 * asynchronously. It is not required to implement both ways, in that case, a
 * UnsupportedOperationException is thrown.
 * 
 * @author dstuerze
 * 
 */
public interface RoutingComponent {

  /** Error message indicating the routing service is not responding */
  public static final int ERROR_SERVICE_DOWN = 1;

  /** Error message indicating the starting point was not found */
  public static final int ERROR_UNKNOWN_START = 2;

  /** Error message indication the destination point was not found */
  public static final int ERROR_UNKNOWN_DESTINATION = 4;

  /** Error message indicating the service could not find a route */
  public static final int ERROR_NO_ROUTE_FOUND = 8;

  /**
   * Calls the routing service synchronously to receive a route between two
   * points. The calculated route will be returned as an array of RoutePoints.
   * If the service does not support synchronous routing, an
   * UnsupportedOperationException will be thrown.
   * 
   * @param from
   *          starting point as GeoCoordinate
   * @param to
   *          target as GeoCoordinate
   * @return RoutePoint[] with route points
   * @throws UnsupportedOperationException
   *           if a synchronous call is not supported
   */
  public RoutePoint[] findRoute(GeoCoordinate from, GeoCoordinate to)
      throws UnsupportedOperationException;

  /**
   * Calls the routing service synchronously to receive a route between two
   * points. The calculated route will be returned as an array of RoutePoints.
   * If the service does not support synchronous routing, an
   * UnsupportedOperationException will be thrown.
   * 
   * @param from
   *          starting point as GeoCoordinate
   * @param to
   *          target as GeoCoordinate
   * @param via
   *          a (possibly empty) set of vias for the route in correct order
   * @return RoutePoint[] with route points
   * @throws UnsupportedOperationException
   *           if a synchronous call is not supported
   */
  public RoutePoint[] findRoute(GeoCoordinate from, GeoCoordinate to,
      GeoCoordinate... via) throws UnsupportedOperationException;

  /**
   * Calls the routing service asynchronously to receive a route between two
   * points. After the route has been calculated, the onRouteCalculated method
   * of the given RoutingListener will be called. If the service does not
   * support asynchronous routing, an UnsupportedOperationException will be
   * thrown.
   * 
   * @param from
   *          starting point as GeoCoordinate
   * @param to
   *          target as GeoCoordinate
   * @param l
   *          the listener to be notified
   * @throws UnsupportedOperationException
   *           if a asynchronous call is not supported
   */
  public void calculateRoute(RoutingListener l, GeoCoordinate from,
      GeoCoordinate to) throws UnsupportedOperationException;

  /**
   * Calls the routing service asynchronously to receive a route between two
   * points. After the route has been calculated, the onRouteCalculated method
   * of the given RoutingListener will be called. If the service does not
   * support asynchronous routing, an UnsupportedOperationException will be
   * thrown.
   * 
   * @param l
   *          the listener to be notified
   * @param from
   *          starting point as GeoCoordinate
   * @param to
   *          target as GeoCoordinate
   * @param via
   *          a (possibly empty) set of vias for the route in correct order
   * @throws UnsupportedOperationException
   *           if a asynchronous call is not supported
   */
  public void calculateRoute(RoutingListener l, GeoCoordinate from,
      GeoCoordinate to, GeoCoordinate... via)
      throws UnsupportedOperationException;
}
