package de.monarchy.guideme.routing.android;

import java.io.IOException;
import java.util.List;

import android.content.Context;
import android.location.Address;
import android.location.Geocoder;
import de.monarchy.guideme.routing.GeoCodingComponent;
import de.monarchy.guideme.routing.GeoCodingListener;
import de.monarchy.guideme.util.GeoCoordinate;
import de.monarchy.guideme.util.PointOfInterest;
import static de.monarchy.guideme.util.GeoCoordinate.*;

/**
 * AndroidGeoCoding is a geocoding provider using Android's internal Geocoder
 * implementation. As described there, it is not defined which service is used.
 * The service is synchronous, so the void method is not supported and throws
 * the corresponding exception.
 * @author dstuerze
 *
 */
public class AndroidGeoCoding implements GeoCodingComponent {

  private static final int MAX_RESULTS = 25;

  private Context context;

  /**
   * Creates a new GeoCoding component with the given application's context.
   * @param x application context
   */
  public AndroidGeoCoding(Context x) {
    context = x;
  }

  @Override
  public PointOfInterest[] findNear(String search, GeoCoordinate location,
      int maxDistance) throws UnsupportedOperationException {
    Geocoder code = new Geocoder(context);
    List<Address> list;
    try {
      if (maxDistance == GeoCodingComponent.EVERYWHERE) {
        // world wide search is easy, as it is just one call
        list = code.getFromLocationName(search, MAX_RESULTS);
      } else {
        // a bounding box is needed
        // distance to north and south in degrees instead of meters
        double nsDeg = maxDistance * 360. / (2 * Math.PI * EARTH_RADIUS);
        double north = location.getLatitude() + nsDeg;
        if (north > 90.)
          north = 90;
        double south = location.getLatitude() - nsDeg;
        if (south < -90.)
          south = -90;
        double latPeri = 2 * Math.PI * EARTH_RADIUS
            * Math.cos(location.getLatitude() * DEG_TO_RAD);
        // distance to east and west in degrees instead of meters
        double ewDeg = maxDistance * 360. / latPeri;
        double west = location.getLongitude() - ewDeg;
        if (west < -180)
          west = west + 360;
        double east = location.getLongitude() + ewDeg;
        if (east > 180)
          east -= 360;
        // call with the bounding box
        list = code.getFromLocationName(search, MAX_RESULTS, south, west,
            north, east);
      }
    } catch (IOException e) {
      e.printStackTrace();
      return new PointOfInterest[0];
    }

    PointOfInterest[] back = new PointOfInterest[list.size()];
    for (int i = 0; i < back.length; i++) {
      // convert the results in pois
      Address a = list.get(i);
      GeoCoordinate g = new GeoCoordinate(a.getLatitude(), a.getLongitude());
      // get the address details and concatenate them
      String feat = a.getFeatureName();
      String locality = a.getLocality();
      String thfare = a.getThoroughfare();
      String plz = a.getPostalCode();
      String desc = "";
      if (thfare != null)
        desc = desc + thfare + " ";
      if (feat != null && !feat.equals(thfare))
        desc = desc + feat + " ";
      if (plz != null)
        desc = desc + plz + " ";
      if (locality != null)
        desc = desc + locality;
      back[i] = new PointOfInterest(g, desc);
    }
    return back;
  }

  /**
   * @throws UnsupportedOperationException in every case
   */
  @Override
  public void findNear(GeoCodingListener l, String search,
      GeoCoordinate location, int maxDistance)
      throws UnsupportedOperationException {
    throw new UnsupportedOperationException(
        "AndroidGeoCoding: no asynchronous geocoding supported");
  }

}
