package de.monarchy.guideme.routing.nutiteq;

import android.util.Log;

import com.nutiteq.components.KmlPlace;
import com.nutiteq.components.WgsPoint;
import com.nutiteq.services.GeocodingResultWaiter;
import com.nutiteq.services.GeocodingService;

import de.monarchy.guideme.data.PersonalLocation;
import de.monarchy.guideme.routing.GeoCodingComponent;
import de.monarchy.guideme.routing.GeoCodingListener;
import de.monarchy.guideme.util.GeoCoordinate;
import de.monarchy.guideme.util.PointOfInterest;

/**
 * NutiteqGeoCoding is a geocoding provider using Nutiteq's Geocoder
 * implementation. The service uses a Cloudmade backend. It is asynchronous, so
 * the method with array return value is not supported and throws the
 * corresponding exception.
 * 
 * @author dstuerze
 * 
 */
public class NutiteqGeoCoding implements GeoCodingComponent {

  private static final int MAX_RESULTS = 25;

  /**
   * @throws UnsupportedOperationException
   *           in every case
   */
  @Override
  public PointOfInterest[] findNear(String search, GeoCoordinate location,
      int maxDistance) throws UnsupportedOperationException {
    throw new UnsupportedOperationException(
        "NutiteqGeoCoding: no synchronous geocoding supported");
  }

  @Override
  public void findNear(GeoCodingListener l, String search,
      GeoCoordinate location, int maxDistance)
      throws UnsupportedOperationException {
    GeoCodingWaiter wait = new GeoCodingWaiter(l, search);
    WgsPoint near = new WgsPoint(location.getLongitude(), location
        .getLatitude());
    new GeocodingService(wait, GeocodingService.DEFAULT_URL,
        PersonalLocation.preferred_language, near, search,
        GeocodingService.SEARCH_TYPE_GEOCODING, new int[0], MAX_RESULTS, false)
        .execute();
  }

  /**
   * Callback for the asynchronous method invocation. The class listens for any
   * results or error the geocoding service is producing.
   * 
   * @author dstuerze
   * 
   */
  private class GeoCodingWaiter implements GeocodingResultWaiter {

    // listener invoking the call
    private GeoCodingListener listener;
    // search string
    private String search;

    /**
     * Builds a new Callback object.
     * 
     * @param l
     *          invoking listener
     * @param s
     *          search string
     */
    private GeoCodingWaiter(GeoCodingListener l, String s) {
      listener = l;
      search = s;
    }

    @Override
    public void errors(int arg0) {
      Log.i("GuideMe", "NutiteqGeoCoding error=" + arg0);
      if (arg0 == GeocodingService.NOT_FOUND)
        listener.onError(ERROR_NOT_FOUND);
      else
        listener.onError(ERROR_SERVICE_DOWN);

    }

    @Override
    public void searchResults(KmlPlace[] arg0) {
      PointOfInterest[] pois = new PointOfInterest[arg0.length];
      if (pois.length == 0) {
        listener.onPlacesFound(search, pois);
        return;
      }
      for (int i = 0; i < pois.length; i++) {
        WgsPoint w = arg0[i].getWgs();
        GeoCoordinate g = new GeoCoordinate(w.getLat(), w.getLon());
        String descr = arg0[i].getName() + " - " + arg0[i].getDescription();
        pois[i] = new PointOfInterest(g, descr);
      }

      listener.onPlacesFound(search, pois);

    }

  }

}
