package de.monarchy.guideme.routing.nutiteq;

import java.util.ArrayList;
import java.util.List;

import android.util.Log;

import com.nutiteq.components.Line;
import com.nutiteq.components.Route;
import com.nutiteq.components.WgsPoint;
import com.nutiteq.services.CloudMadeDirections;
import com.nutiteq.services.DirectionsWaiter;

import de.monarchy.guideme.routing.RoutingComponent;
import de.monarchy.guideme.routing.RoutingListener;
import de.monarchy.guideme.util.GeoCoordinate;
import de.monarchy.guideme.util.RoutePoint;
import static de.monarchy.guideme.util.GeoCoordinate.*;

/**
 * Implements a RoutingComponent for the nutiteq maps api. It uses Cloudmade
 * asynchronous routing as routing service.
 * 
 * @author dstuerze
 * 
 */
public class NutiteqRouting implements RoutingComponent {

  // Cloudmade api key
  private static final String apiKey = "d23f55008e5949869a7d0b15e45f0b20";

  @Override
  public void calculateRoute(RoutingListener l, GeoCoordinate from,
      GeoCoordinate to) throws UnsupportedOperationException {
    // convert start and end position
    WgsPoint start = new WgsPoint(from.getLongitude(), from.getLatitude());
    WgsPoint dest = new WgsPoint(to.getLongitude(), to.getLatitude());
    CloudmadeWaiter waiter = new CloudmadeWaiter(l);
    // execute routing
    new CloudMadeDirections(waiter, start, dest,
        CloudMadeDirections.ROUTE_TYPE_FOOT, apiKey).execute();
    Log.d("GuideMe", "NutiteqRouting routing dest=" + to.toString());
  }

  /**
   * @throws UnsupportedOperationException
   *           in every case
   */
  @Override
  public void calculateRoute(RoutingListener l, GeoCoordinate from,
      GeoCoordinate to, GeoCoordinate... via)
      throws UnsupportedOperationException {
    throw new UnsupportedOperationException(
        "NutiteqRouting: no via routing supported");

  }

  /**
   * @throws UnsupportedOperationException
   *           in every case
   */
  @Override
  public RoutePoint[] findRoute(GeoCoordinate from, GeoCoordinate to)
      throws UnsupportedOperationException {
    throw new UnsupportedOperationException(
        "NutiteqRouting: no synchronous routing supported");
  }

  /**
   * @throws UnsupportedOperationException
   *           in every case
   */
  @Override
  public RoutePoint[] findRoute(GeoCoordinate from, GeoCoordinate to,
      GeoCoordinate... via) throws UnsupportedOperationException {
    throw new UnsupportedOperationException(
        "NutiteqRouting: no synchronous routing supported");
  }

  /**
   * Callback for the asynchronous method invocation. The class listens for any
   * results or error the routing service is producing.
   * 
   * @author dstuerze
   * 
   */
  private class CloudmadeWaiter implements DirectionsWaiter {

    // invoking listener
    private RoutingListener listener;

    /**
     * Creates new callback for one listener.
     * 
     * @param l
     *          invoking listener
     */
    private CloudmadeWaiter(RoutingListener l) {
      listener = l;
    }

    @Override
    public void networkError() {
      listener.onError(ERROR_SERVICE_DOWN);
    }

    @Override
    public void routeFound(Route arg0) {
      // instructions by nutiteq
      Line l = arg0.getRouteLine();
      WgsPoint[] p = l.getPoints();
      // wrapper
      List<RoutePoint> route = new ArrayList<RoutePoint>();

      // actual position
      GeoCoordinate gc = null;
      // has to be created beforehand (it's cheaper)
      if (p.length > 0) {
        WgsPoint pos = p[0];
        gc = new GeoCoordinate(pos.getLat(), pos.getLon());
      }
      // track angle diff
      double tA = -1000.;
      // creating all those route point without the last one
      for (int i = 0; i < p.length - 1; i++) {
        // next position
        WgsPoint pos = p[i + 1];
        GeoCoordinate vgl = new GeoCoordinate(pos.getLat(), pos.getLon());
        // distance from here (gc) to next (vgl)
        int dist = (int) (gc.distanceTo(vgl) * EARTH_RADIUS);
        // track angle
        double nTA = gc.trackAngle(vgl) * RAD_TO_DEG;
        Log.i("GuideMe", "nTA=" + nTA);
        // wrapping the information
        if (Math.abs(nTA - tA) >= 5.) {
          route.add(new RoutePoint(gc, "", dist));
          tA = nTA;
        }
        // save time: the next point will shortly be the actual
        gc = vgl;
      }

      // computing the last position: this one has no distance
      // String instr = inst[inst.length - 1].getInstruction();
      route.add(new RoutePoint(gc, "", 0));

      // tell the listener what you know
      listener.onRouteCalculated(route.toArray(new RoutePoint[route.size()]));
    }

    @Override
    public void routingErrors(int arg0) {
      if (arg0 == com.nutiteq.services.DirectionsService.ERROR_DESTINATION_ADDRESS_NOT_FOUND)
        listener.onError(ERROR_UNKNOWN_DESTINATION);
      if (arg0 == com.nutiteq.services.DirectionsService.ERROR_FROM_ADDRESS_NOT_FOUND)
        listener.onError(ERROR_UNKNOWN_START);
      if (arg0 == com.nutiteq.services.DirectionsService.ERROR_ROUTE_NOT_FOUND
          || arg0 == com.nutiteq.services.DirectionsService.ERROR_FROM_AND_DESTINATION_ADDRESS_SAME)
        listener.onError(ERROR_NO_ROUTE_FOUND);
    }

    @Override
    public void routingParsingError(String arg0) {
      listener.onError(ERROR_NO_ROUTE_FOUND);
    }
  }
}