package de.monarchy.guideme.routing.osm;

import java.io.IOException;
import java.net.MalformedURLException;
import java.net.URL;

import javax.xml.parsers.ParserConfigurationException;
import javax.xml.parsers.SAXParser;
import javax.xml.parsers.SAXParserFactory;

import org.xml.sax.InputSource;
import org.xml.sax.SAXException;
import org.xml.sax.XMLReader;

import android.util.Log;
import de.monarchy.guideme.routing.GeoCodingComponent;
import de.monarchy.guideme.routing.GeoCodingListener;
import de.monarchy.guideme.util.GeoCoordinate;
import de.monarchy.guideme.util.PointOfInterest;

/**
 * OSMGeoCoding is a geocoding provider using OpenStreetMap's Geocoder
 * implementation "Namefinder". The service is synchronous, so the void method
 * is not supported and throws the corresponding exception.
 * 
 * @author dstuerze
 * 
 */
public class OSMGeoCoding implements GeoCodingComponent {

  @Override
  public PointOfInterest[] findNear(String search, GeoCoordinate location,
      int maxDistance) throws UnsupportedOperationException {
    // getting an xml parser
    SAXParserFactory factory = SAXParserFactory.newInstance();
    XMLReader xml;
    try {
      SAXParser parser = factory.newSAXParser();
      xml = parser.getXMLReader();
    } catch (SAXException e) {
      // as all following: if something goes wrong, just
      // return an empty list
      Log.i("GuideMe", "OSMGeoCoding sax parser error");
      return new PointOfInterest[0];
    } catch (ParserConfigurationException e) {
      Log.i("GuideMe", "OSMGeoCoding sax parser error");
      return new PointOfInterest[0];
    }
    // content handler implementation
    OSMGeoHandler x = new OSMGeoHandler(location, maxDistance);
    xml.setContentHandler(x);

    // webservice url
    URL url = parseUrl(location, search);
    if (url == null)
      return new PointOfInterest[0];
    InputSource input = new InputSource();
    try {
      // open the stream and parse
      Log.d("GuideMe", "OSMGeoCoding call: " + url.toString());
      input.setByteStream(url.openStream());
      xml.parse(input);
      return x.getEntries();
    } catch (IOException e) {
      Log.i("GuideMe", "OSMGeoCoding xml stream reading error");
      return new PointOfInterest[0];
    } catch (SAXException e) {
      Log.i("GuideMe", "OSMGeoCoding sax parser error");
      return new PointOfInterest[0];
    }
  }

  /**
   * @throws UnsupportedOperationException
   *           in every case
   */
  @Override
  public void findNear(GeoCodingListener l, String search,
      GeoCoordinate location, int maxDistance)
      throws UnsupportedOperationException {
    throw new UnsupportedOperationException(
        "OSMGeoCoding: no asynchronous geocoding supported");
  }

  /**
   * Creates a webservice url for a specific location and radius.
   * 
   * @param point
   *          central position
   * @param radius
   *          radius around the position (in metres)
   * @return webservice url or null, if an error occured
   */
  private URL parseUrl(GeoCoordinate point, String search) {
    StringBuffer url = new StringBuffer();
    // base domain
    url
        .append("http://gazetteer.openstreetmap.org/namefinder/search.xml?find=");
    // search string
    url.append(search);
    // near location
    String loc = " near " + point.getLatitude() + "," + point.getLongitude();
    url.append(loc);
    // maximum number of locations- not too many
    url.append("&max=20");
    try {
      URL wUrl = new URL(url.toString());
      return wUrl;
    } catch (MalformedURLException e) {
      Log.i("GuideMe", "OSMGeoCoding bad url");
      return null;
    }
  }

}
