package de.monarchy.guideme.routing.osm;

import java.util.HashSet;
import java.util.Set;

import org.xml.sax.Attributes;
import org.xml.sax.ContentHandler;
import org.xml.sax.Locator;
import org.xml.sax.SAXException;

import de.monarchy.guideme.routing.GeoCodingComponent;
import de.monarchy.guideme.util.GeoCoordinate;
import de.monarchy.guideme.util.PointOfInterest;

/**
 * This class is a content handler for sax xml parsing. It is specialised for
 * the openstreetmap geocoding web service.
 * 
 * @author dstuerze
 * 
 */
public class OSMGeoHandler implements ContentHandler {

  // all those locations found
  private Set<PointOfInterest> articles = new HashSet<PointOfInterest>();
  // filter attributes 
  private int depth = 0;
  private int maxDistance = 0;
  private GeoCoordinate pos;

  /**
   * Creates the handler for the geocoder's xml stream. It needs the data provided,
   * as it does not use delegation to filter out locations. 
   * @param where position around which is coded
   * @param radius maximum distance to the location
   */
  public OSMGeoHandler(GeoCoordinate where, int radius) {
    pos = where;
    maxDistance = radius;
  }

  /**
   * Returns the entries found.
   * 
   * @return all articles as array
   */
  public PointOfInterest[] getEntries() {
    return articles.toArray(new PointOfInterest[0]);
  }

  @Override
  public void startElement(String uri, String localName, String qName,
      Attributes atts) throws SAXException {
    if (localName.equals("named")) {
      // named means: new location
      if (depth == 0) {
        // there are additional locations for each interesting location,
        // so only look at top level
        
        // get element's necessary attributes
        String name = atts.getValue("", "name");
        double lat = Double.parseDouble(atts.getValue("", "lat"));
        double lon = Double.parseDouble(atts.getValue("", "lon"));
        GeoCoordinate gc = new GeoCoordinate(lat, lon);
        if (pos.distanceTo(gc) * GeoCoordinate.EARTH_RADIUS <= maxDistance
            || maxDistance == GeoCodingComponent.EVERYWHERE) {
          // if the location is within radius, take it
          PointOfInterest put = new PointOfInterest(gc, name);
          articles.add(put);
        }
      }
      // next named will be a level deeper
      depth++;

    }
  }

  @Override
  public void endElement(String uri, String localName, String qName)
      throws SAXException {
    if (localName.equals("named")) {
      // next named will be a level higher
      depth--;
    }
  }

  // What follows are empty methods!
  @Override
  public void endDocument() throws SAXException {
  }

  @Override
  public void startDocument() throws SAXException {
  }

  @Override
  public void endPrefixMapping(String prefix) throws SAXException {
  }

  @Override
  public void ignorableWhitespace(char[] ch, int start, int length)
      throws SAXException {
  }

  @Override
  public void processingInstruction(String target, String data)
      throws SAXException {
  }

  @Override
  public void setDocumentLocator(Locator locator) {
  }

  @Override
  public void skippedEntity(String name) throws SAXException {
  }

  @Override
  public void startPrefixMapping(String prefix, String uri) throws SAXException {
  }

  @Override
  public void characters(char[] ch, int start, int length) throws SAXException {
  }

}
