package de.monarchy.guideme.util;

import java.io.Serializable;

/**
 * GeoCoordinate represents a location on a globe, represented by latitude and
 * longitude. Latitude is measured in degrees from -90 to 90, with 0 being at
 * the equator and positive values representing positions on the northern
 * hemisphere, negative values positions on the southern. Longitude is measured
 * in degrees from -180 to 180 with 0 being the prime meridian and negative
 * values being west of it, positive in the east.
 * 
 * @author dstuerze
 * 
 */
public class GeoCoordinate implements Serializable {

  /** */
  private static final long serialVersionUID = 897195823616852473L;

  /** earth radius in meters */
  public static final double EARTH_RADIUS = 6371010.;

  /** conversion of degrees to radian */
  public static final double DEG_TO_RAD = Math.PI / 180.;

  /** conversion of radian to degrees */
  public static final double RAD_TO_DEG = 180. / Math.PI;

  private double longitude;
  private double latitude;

  /**
   * Creates a geo coordinate representing the position at lat, lon
   * 
   * @param lat
   *          latitude
   * @param lon
   *          longitude
   */
  public GeoCoordinate(double lat, double lon) {
    setLatitude(lat);
    setLongitude(lon);
  }

  /**
   * Copies the values of the given coordinate to the new one.
   * 
   * @param copy
   *          coordinate to be copied
   */
  public GeoCoordinate(GeoCoordinate copy) {
    setLatitude(copy.getLatitude());
    setLongitude(copy.getLongitude());
  }

  /**
   * Returns the longitude.
   * 
   * @return longitude
   */
  public double getLongitude() {
    return longitude;
  }

  /**
   * Sets the longitude. Value has to be between -180 and 180.
   * 
   * @param longitude
   *          new longitude
   * @throws IllegalArgumentException
   *           if values unfitting
   */
  public void setLongitude(double longitude) {
    if (longitude < -180 || longitude > 180)
      throw new IllegalArgumentException("illegal longitude " + longitude);

    this.longitude = longitude;
  }

  /**
   * Returns the latitude.
   * 
   * @return latitude
   */
  public double getLatitude() {
    return latitude;
  }

  /**
   * Sets the latitude. Value has to be between -90 and 90.
   * 
   * @param latitude
   *          new latitude
   * @throws IllegalArgumentException
   *           if values unfitting
   */
  public void setLatitude(double latitude) {
    if (latitude < -90 || latitude > 90)
      throw new IllegalArgumentException("illegal latitude " + latitude);
    this.latitude = latitude;
  }

  /**
   * Returns the distance between to points on the unit sphere. The sphere in
   * this case has a radius of 1 metre, so the result has to be scaled if
   * necessary.
   * 
   * The distance calculated is the great-circle distance between two points,
   * which means it represents the shortest possible path.
   * 
   * As for performance reasons, the distance is not calculated in wgs 84, but
   * on a sphere.
   * 
   * @param another
   *          other point the distance is calculated to
   * @return distance on unit sphere
   */
  public double distanceTo(GeoCoordinate another) {
    // transforming coordinates to radian
    double latthis = this.latitude * DEG_TO_RAD;
    double lonthis = this.longitude * DEG_TO_RAD;
    double latother = another.latitude * DEG_TO_RAD;
    double lonother = another.longitude * DEG_TO_RAD;

    // calculating the distance in four steps
    double sinpart = Math.sin(latthis) * Math.sin(latother);
    double cospart = Math.cos(latthis) * Math.cos(latother);
    cospart = cospart * Math.cos(lonother - lonthis);
    return Math.acos(sinpart + cospart);
  }

  /**
   * Returns the angle a given point lies in relation to this. The track angle
   * is zero at north, increasing clockwise in radian.
   * 
   * This method uses rhumb lines to calculate the direction, so the angle will
   * remain the same while the objects would approach each other.
   * 
   * Movement along a rhumb line is not the shortest way. At distances of up to
   * 30 degrees in longitude the bias is less than one percent.
   * 
   * The rhumb line is calculated on a unit sphere for performance reasons.
   * 
   * @param another
   *          other point whose direction is needed
   * @return angle in radian
   */
  public double trackAngle(GeoCoordinate another) {
    // transforming coordinates to radian
    double lonthis = this.longitude * DEG_TO_RAD;
    double lonother = another.longitude * DEG_TO_RAD;

    // doing the mercator projection
    double ythis = Math.log(Math.tan(this.latitude * DEG_TO_RAD / 2. + Math.PI
        / 4.));
    double yother = Math.log(Math.tan(another.latitude * DEG_TO_RAD / 2.
        + Math.PI / 4.));
    double atan = Math.atan2(yother - ythis, lonother - lonthis);
    double angle = Math.PI / 2. - atan;
    if (angle < 0)
      angle = angle + 2 * Math.PI;

    // so far, we always go east, but maybe it's shorter the other way around
    return (Math.abs(lonother - lonthis) > Math.PI) ? (Math.PI + angle)
        % (2 * Math.PI) : angle % (2 * Math.PI);
  }

  public boolean equals(Object o) {
    if (o == null)
      return false;
    if (o.getClass() != this.getClass())
      return false;
    GeoCoordinate g = (GeoCoordinate) o;
    return latitude == g.latitude && longitude == g.longitude;
  }

  public int hashCode() {
    Double lo = longitude;
    Double la = latitude;
    return lo.hashCode() + la.hashCode();
  }

  public String toString() {
    return "(" + latitude + ", " + longitude + ")";
  }

}
