package de.monarchy.guideme.util;

import android.content.Context;
import android.view.ContextMenu;
import android.view.View;


/**
 * Specifies a location with a special name. The class PointOfInterest will
 * be superclass of all points of interest, as its name says. The class itself
 * can already be used as a simple location.
 *  
 * @author dstuerze
 *
 */
public class PointOfInterest extends GeoCoordinate implements POI, ContextMenu.ContextMenuInfo {
  
  /** */
  private static final long serialVersionUID = 4239881782795611864L;
  
  /** Description of the location */
  private final String description;

  /**
   * Instanciates a PointOfInterest at the given location with the specified
   * name.
   *  
   * @param lat latitude of the POI
   * @param lon longitude of the POI
   * @param descr short description
   */
  public PointOfInterest(double lat, double lon, String descr) {
    super(lat, lon);
    description = descr;
  }
  
  /**
   * Instanciates a PointOfInterest at the given location with the specified
   * name.
   * 
   * @param g location of the POI
   * @param descr short description
   */
  public PointOfInterest(GeoCoordinate g, String descr) {
    this(g.getLatitude(), g.getLongitude(), descr);
  }
  
  public String getDescription() {
    return description;
  }
  
  /**
   * Returns the location as GeoCoordinate. Although a PointOfInterest is 
   * already a subclass of GeoCoordinate, it will return a new instance
   * of GeoCoordinate to hide its own informations and to prevent side
   * effects.
   * 
   * @return GeoCoordinate holding location information
   */
  public GeoCoordinate getLocation() {
    return new GeoCoordinate(this);
  }
  
  /**
   * A general PointOfInterest has no specified view. Subclasses should override
   * this method.  
   * @param c context for the view
   * @return the view
   */
  public View getView(Context c) {
    return null;
  }
  
  public boolean equals(Object o) {
    if (o == null)
      return false;
    if (o.getClass() != this.getClass())
      return false;
    PointOfInterest p = (PointOfInterest)o;
    if (!getLocation().equals(p.getLocation()))
      return false;
    return getDescription().equals(p.getDescription());
    
  }
  
  public int hashCode() {
    return getLocation().hashCode() + description.hashCode();
  }

}
