package de.monarchy.guideme.view;

import java.io.IOException;

import android.content.Context;
import android.hardware.Camera;
import android.util.Log;
import android.view.SurfaceHolder;
import android.view.SurfaceView;

/**
 * CameraSurface is a SurfaceView able to hold the camera preview. It is mainly based
 * on a former Android Developer guide article.
 * 
 * 
 * @author dstuerze
 *
 */
public class CameraSurface extends SurfaceView implements
    SurfaceHolder.Callback {
  private SurfaceHolder holder;
  private Camera camera;
  
  private static CameraSurface instance;

  /**
   * Creates a surface for the given context.
   * @param context application context
   */
  private CameraSurface(Context context) {
    super(context);
    Log.i("GuideMe", "CameraSurface instance");
    
    // access to the surface is possible via the Holder
    holder = getHolder();
    // information will come from elsewhere
    holder.setType(SurfaceHolder.SURFACE_TYPE_PUSH_BUFFERS);
    // to get information about the view, we have to add a callback
    holder.addCallback(this);
    
  }

  public void surfaceCreated(SurfaceHolder holder) {
    // surface is visible, so get the camera
    try {
      Log.i("GuideMe", "CameraSurface surfaceCreated");
      
      camera = Camera.open();
      // tell the cam where to put a preview
      camera.setPreviewDisplay(holder);
    } catch (IOException exception) {
      // if the cam isn't able to draw there, we have to give it away
      Log.w("GuideMe", "CameraSurface IOException " + exception.getMessage());
      camera.release();
      camera = null;
    } catch (RuntimeException e) {
      Log.w("GuideMe", "CameraSurface exception ");
      if (camera != null)
        camera.release();
      camera = null;
      e.printStackTrace();
    }
  }

  public void surfaceDestroyed(SurfaceHolder holder) {
    // if the surface isn't visible, the preview has to stop
    // else nobody will be able to use it
    if (camera != null) {
      camera.stopPreview();
      camera.release();
      camera = null;
    }
    // as the preview image was buffered, we should clean up a bit
    System.gc();
  }

  public void surfaceChanged(SurfaceHolder holder, int format, int w, int h) {
    // if something changed with the surface, we redo the camera
    System.gc();
    if (camera != null) {
      camera.stopPreview();
      Camera.Parameters parameters = camera.getParameters();
      parameters.setPreviewSize(w, h);
      camera.setParameters(parameters);
      camera.startPreview();
    }
  }
  
  /**
   * Returns an instance of the CameraSurface. CameraSurface is a singleton,
   * so it only creates one instance.
   * 
   * @param x
   *          application context
   * @return the one and only instance
   */
  public static CameraSurface getInstance(Context x) {
    if (instance == null)
      instance = new CameraSurface(x);
    return instance;
  }

}
