package de.monarchy.guideme.view;

import android.app.Activity;
import android.view.View;
import android.view.ViewGroup;
import android.widget.AbsoluteLayout;
import android.widget.RelativeLayout;
import de.monarchy.guideme.main.OrientationChangeListener;
import de.monarchy.guideme.main.PositionChangeListener;
import de.monarchy.guideme.main.RoutingListener;
import de.monarchy.guideme.map.MapWrapper;
import de.monarchy.guideme.map.nutiteq.NutiteqWrapper;
import de.monarchy.guideme.util.GeoCoordinate;
import de.monarchy.guideme.util.POI;
import de.monarchy.guideme.util.RoutePoint;

/**
 * The CombinationView is the linking view of augmention and reality. It
 * combines camera, POI views, direction arrow and map view and is responsible
 * for a correct interaction between them.
 * 
 * @author dstuerze
 * 
 */
@SuppressWarnings("deprecation")
public class CombinationView extends AbsoluteLayout implements
    OrientationChangeListener, PositionChangeListener, RoutingListener {

  // the views
  private RealityView cam;
  private View mapView;
  private FlatArrow flatArrow;

  // mapping controller
  private MapWrapper map;

  // device angles
  private float zAngle;
  private float yAngle;

  // activation codes of the views
  private boolean isMapShowing;
  private boolean mapActivated;
  private boolean isCamShowing;
  private boolean camActivated;

  // filter variable for angle changes
  private static final float E = 2f;

  /** Roll below which the camera is deactivated */
  public static final int CAM_THRESHOLD = 25;
  /** Roll above which the map is deactivated */
  public static final int MAP_THRESHOLD = 70;

  /**
   * Creating a combination view for the given context. The necessary view will
   * be instanciated and ordered.
   * 
   * @param context
   *          application context
   */
  public CombinationView(Activity context) {
    super(context);

    // map controller
    map = NutiteqWrapper.getInstance(context);

    // creating views
    cam = new RealityView(context);
    mapView = map.getView();
    flatArrow = FlatArrow.getInstance(context);

    // adding views
    this.addView(cam, new LayoutParams(ViewGroup.LayoutParams.FILL_PARENT,
        ViewGroup.LayoutParams.FILL_PARENT, 0, 0));

    isCamShowing = true;
    isMapShowing = true;
    mapActivated = true;
    camActivated = true;
  }

  @Override
  public void stateChanged(float[] arg) {
    // device is rolled
    float newAngle = arg[2];
    if (Math.abs(zAngle - newAngle) > E) {
      // if a modification was made
      zAngle = newAngle;

      if ((newAngle < CAM_THRESHOLD && isCamShowing && mapActivated)
          || !camActivated) {
        // device too flat for cam or cam deactivated, so deactivate view
        cam.setVisibility(INVISIBLE);
        isCamShowing = false;
        map.setFullScreen(true);
      }
      if (newAngle >= CAM_THRESHOLD && !isCamShowing && camActivated) {
        // device steep enough for cam and cam activated, so activate view
        cam.setVisibility(VISIBLE);
        isCamShowing = true;
        map.setFullScreen(false);
      }
      if ((newAngle > MAP_THRESHOLD && isMapShowing && camActivated)
          || !mapActivated) {
        // device too steep for map or map deactivated, so deactivate view
        mapView.setVisibility(INVISIBLE);
        isMapShowing = false;
        map.setDrawing(false);
      }
      if (newAngle <= MAP_THRESHOLD && mapActivated) {
        if (!isMapShowing) {
          // device flatenough for map and map activated, so activate view
          mapView.setVisibility(VISIBLE);
          isMapShowing = true;
          map.setDrawing(true);
        }

        // linearly change the size of the map
        newAngle = ((MAP_THRESHOLD - CAM_THRESHOLD) - (newAngle - CAM_THRESHOLD))
            / (MAP_THRESHOLD - CAM_THRESHOLD);
        int h = getHeight();
        int myH;
        if (arg[2] < CAM_THRESHOLD || !camActivated)
          // show the map fullscreen
          myH = h;
        else
          // just show a piece
          myH = (int) (h * newAngle);
        LayoutParams mapp = new LayoutParams(
            ViewGroup.LayoutParams.FILL_PARENT,
            ViewGroup.LayoutParams.FILL_PARENT, 0, h - myH);
        mapView.setLayoutParams(mapp);
        // tell the map for panning
        map.onOrientationChanged(arg);
      }
      invalidate();
    }
    if (Math.abs(yAngle - arg[0]) > E) {
      // if the view angle changed, call the views (if they are active)
      yAngle = arg[0];
      if (isMapShowing)
        map.onOrientationChanged(arg);
      if (isCamShowing) {
        cam.onOrientationChanged(arg);
      }
      flatArrow.setOrientation(arg[0]);
    }
  }

  @Override
  public void stateChanged(GeoCoordinate arg0, POI[] arg1) {
    // call the views, if the position changed, a layout check is not necessary
    map.onPositionChanged(arg0, arg1);
    cam.onPositionChanged(arg0, arg1);
    flatArrow.setLocation(arg0);

  }

  /**
   * Activate or deactivate the map view. If the map is deactivated, it will not
   * be shown.
   * 
   * @param b
   *          new status
   */
  public void setMapActivated(boolean b) {
    mapActivated = b;
  }

  /**
   * Returns whether the map view is currently activated. This does not check
   * whether the view is visible or invisible because of the current roll of the
   * device.
   * 
   * @return true if the map is activated, false else
   */
  public boolean isMapActivated() {
    return mapActivated;
  }

  /**
   * Activate or deactivate the cam view. If the cam is deactivated, it will not
   * be shown.
   * 
   * @param b
   *          new status
   */
  public void setCamActivated(boolean b) {
    camActivated = b;
  }

  /**
   * Returns whether the camera view is currently activated. This does not check
   * whether the view is visible or invisible because of the current roll of the
   * device.
   * 
   * @return true if the cam is activated, false else
   */
  public boolean isCamActivated() {
    return camActivated;
  }

  /**
   * Callback method for when the view is invisible. It is necessary to release
   * the map view as only one is existing and it might be needed elsewhere.
   */
  public void onStop() {
    cam.onStop();
    removeView(mapView);
    removeView(flatArrow);
  }

  /**
   * Callback method for when the view is becoming visible. The map view will be
   * added to the layout and its appearance will be adjusted.
   */
  public void onStart() {
    LayoutParams par = new LayoutParams(ViewGroup.LayoutParams.FILL_PARENT,
        ViewGroup.LayoutParams.FILL_PARENT, 0, 0);
    addView(mapView, par);
    addView(flatArrow, new LayoutParams(120, 120, 190, 170));
    // the map is fixed, shows a view angle and has zoom controls in the upper
    // right corner
    map.setScrollable(false);
    map.showViewDirection(true);
    map.positionZoomControls(RelativeLayout.ALIGN_PARENT_RIGHT,
        RelativeLayout.ALIGN_PARENT_TOP);
    cam.onStart();
  }

  /**
   * Callback method for when a route is deleted. The information is forwarded
   * to the interested views, here map and direction arrow.
   */
  public void removeRoute() {
    map.removeRoute();
    flatArrow.removeRoute();
  }

  @Override
  public void routingDone(RoutePoint[] points) {
    if (points != null && points.length > 0) {
      map.setRoute(points);
      flatArrow.setRoute(points);
    }
  }
}
