package de.monarchy.guideme.view;

import static de.monarchy.guideme.util.GeoCoordinate.EARTH_RADIUS;
import android.content.Context;
import android.graphics.Bitmap;
import android.graphics.BitmapFactory;
import android.graphics.Canvas;
import android.graphics.CornerPathEffect;
import android.graphics.Paint;
import android.graphics.Path;
import android.graphics.PorterDuff;
import android.graphics.PorterDuffColorFilter;
import android.graphics.Rect;
import android.util.Log;
import android.view.View;
import de.monarchy.guideme.R;
import de.monarchy.guideme.util.GeoCoordinate;
import de.monarchy.guideme.util.RoutePoint;

/**
 * This view represents an arrow indicating the direction the user has to turn
 * to. The arrow is only displayed when a route has been set. It has two color
 * indicators, one for the distance to the destination, one for the distance to
 * the next route point. The first is the fill color of the arrow, the latter is
 * the border color. The arrow reacts to both orientation and position changes
 * and calculates the next route point autonomously.
 * 
 * @author dstuerze
 * 
 */
public class FlatArrow extends View {

  // minimum change of orientation in degrees
  private static final int E = 9;

  // minimum and maximum distance to target for color change
  private static final int MIN_DIST = 20;
  private static final int SHORT_MAX_DIST = 100;
  private static final int LONG_MAX_DIST = 783;

  // color indicating dist to next turning point
  private static Paint strokePaint;
  // color indicating dist to destination
  private static Paint fillPaint;
  // target flag
  private static Bitmap flag;

  static {
    // initial paint style object for arrow filling
    fillPaint = new Paint();
    fillPaint.setAntiAlias(true);
    fillPaint.setARGB(0xC0, 0x40, 0xFF, 0);
    fillPaint.setStyle(Paint.Style.FILL);
    fillPaint.setPathEffect(new CornerPathEffect(3));

    // initial paint style object for arrow border
    strokePaint = new Paint();
    strokePaint.setAntiAlias(true);
    strokePaint.setARGB(0xFF, 0xFF, 0x40, 0);
    strokePaint.setStyle(Paint.Style.STROKE);
    strokePaint.setPathEffect(new CornerPathEffect(1));
    strokePaint.setStrokeWidth(4);
  }

  // angles to target and view
  private float trackAngle, viewAngle;
  // actual and desired position
  private GeoCoordinate location, target;
  // route
  private RoutePoint[] route;
  // next route point
  private int next;

  // singleton instance
  private static FlatArrow instance;

  /**
   * Creates a new Arrow for the given application context. As it has no initial
   * route, it will not be shown.
   * 
   * @param context
   *          application context
   */
  private FlatArrow(Context context) {
    super(context);
    // destination symbol is static, so fill only once
    if (flag == null)
      flag = BitmapFactory.decodeResource(context.getResources(),
          R.drawable.target);
  }

  /**
   * Returns an instance of the FlatArrow. FlatArrow is a singleton, so it only
   * creates one instance.
   * 
   * @param context
   *          application context
   * @return the one and only instance
   */
  public static FlatArrow getInstance(Context context) {
    if (instance == null)
      instance = new FlatArrow(context);
    return instance;
  }

  public void onDraw(Canvas c) {
    // don't draw if no target is set
    if (target == null)
      return;
    // saving rotation matrix
    c.save();
    // get view height and width
    int w = getWidth();
    int h = getHeight();
    // create arrow
    Path p = new Path();
    if ((next == route.length - 1 && location.distanceTo(target) * EARTH_RADIUS < MIN_DIST)) {
      // case destination reached: draw flag
      Rect dst = new Rect(0, 0, getLayoutParams().height,
          getLayoutParams().width);
      Paint col = new Paint();
      // draw it green
      col.setColorFilter(new PorterDuffColorFilter(0xFF00FF00, PorterDuff.Mode
          .valueOf("SRC_ATOP")));
      c.drawBitmap(flag, null, dst, col);
    } else {
      if ((720 - viewAngle + trackAngle) % 360 > 90
          && (720 - viewAngle + trackAngle) % 360 < 270) {
        // case target is behind the user
        p.moveTo(w * 6 / 12, h * 3 / 12);
        p.rLineTo(-w * 4 / 12, h * 5 / 12);
        p.rLineTo(w * 2 / 12, 0);
        p.rLineTo(w / 12, h * 1 / 12);
        p.rLineTo(w * 2 / 12, 0);
        p.rLineTo(w / 12, -h * 1 / 12);
        p.rLineTo(w * 2 / 12, 0);
        p.close();
      } else {
        // case target is in front of user
        p.moveTo(w / 2, h * 2 / 12);
        p.rLineTo(-w * 4 / 12, h * 2 / 12);
        p.rLineTo(w * 2 / 12, 0);
        p.rLineTo(-w * 3 / 12, h * 4 / 12);
        p.rLineTo(w * 10 / 12, 0);
        p.rLineTo(-w * 3 / 12, -h * 4 / 12);
        p.rLineTo(w * 2 / 12, 0);
        p.close();
      }
      // draw in corect direction by rotating the view
      c.rotate(360 - viewAngle + trackAngle, w / 2, h * 6 / 12);
      // arrow filling
      c.drawPath(p, fillPaint);
      // arrow bounds
      c.drawPath(p, strokePaint);
    }
    // restore rotation matrix
    c.restore();
  }

  /**
   * Returns the current location of the arrow.
   * 
   * @return location
   */
  public GeoCoordinate getLocation() {
    return location;
  }

  /**
   * Sets the location and redraws. Repainting is only done if target is set.
   * 
   * @param location
   *          new location
   */
  public void setLocation(GeoCoordinate location) {
    // set location
    this.location = location;
    if (target != null) {
      // calculate stage
      setNextRoutePoint();
      // calculate angle to target
      // calculate distance color indicators
      setFillColor();
      setStrokeColor();
      invalidate();
    }
  }

  /**
   * Sets a new route to the arrow. If null is passed, the route will be deleted
   * and the arrow will not be drawn anymore.
   * 
   * @param way
   *          new route
   */
  public void setRoute(RoutePoint[] way) {
    // assign route
    route = way;
    next = -1;
    // route deleted? else, get destination
    target = way != null ? way[way.length - 1] : null;
    if (target != null) {
      // next step
      next = 0;
      // maybe we don't have a location, so assume you start at the start
      if (location == null)
        location = route[0];
      // do the usual calculations
      setNextRoutePoint();
    }
    setFillColor();
    setStrokeColor();
    invalidate();
  }

  /**
   * Removes the route. Convenience method for setRoute(null).
   */
  public void removeRoute() {
    // delete route and unpaint
    route = null;
    target = null;
    next = -1;
    invalidate();
  }

  /**
   * Sets a new oritentation. The orientation has to be in degrees, with 0 being
   * north, incrementing clockwise (east 90 degrees, south 180, west 270). The
   * orientation has to be delivered in device coordinates, the rotation of 90
   * degrees is measured inside the method.
   * 
   * @param angle
   *          new orientation
   */
  public void setOrientation(float angle) {
    if (Math.abs(viewAngle - (angle + 90) % 360) > E) {
      viewAngle = (angle + 90) % 360;
      invalidate();
    }
  }

  /**
   * Calculates the fill color of the arrow. Between MIN_DIST and MAX_DIST the
   * color linearly changes from bright green to bright red.
   */
  private void setFillColor() {
    double dist = location.distanceTo(route[next]) * EARTH_RADIUS;
    for (int i = next; i < route.length - 1; i++) {
      dist = dist + route[i].distanceTo(route[i + 1]) * EARTH_RADIUS;
    }
    int green;
    int red;
    if (dist < (LONG_MAX_DIST + MIN_DIST) / 2.) {
      // from green to yellow
      green = 255;
      double aDist = Math.max(dist - MIN_DIST, 0);
      red = (int) (Math.max(0,
          255. * (aDist / ((LONG_MAX_DIST + MIN_DIST) / 2.))));
    } else {
      // from yellow to red
      red = 255;
      double aDist = Math.min((LONG_MAX_DIST + MIN_DIST) / 2., dist
          - (LONG_MAX_DIST + MIN_DIST) / 2.);
      green = (int) (Math.max(0,
          255 - 255. * (aDist / ((LONG_MAX_DIST + MIN_DIST) / 2.))));
    }
    Log.v("GuideMe", "FlatArrow fillColor: r=" + red + ",g=" + green + ",d="
        + dist);
    fillPaint.setARGB(0xC0, red, green, 0);
  }

  /**
   * Calculates the border color of the arrow. Between MIN_DIST and MAX_DIST the
   * color linearly changes from bright green to bright red.
   */
  private void setStrokeColor() {
    double dist = location.distanceTo(route[next]) * EARTH_RADIUS;
    int green;
    int red;
    if (dist < (SHORT_MAX_DIST + MIN_DIST) / 2.) {
      green = 255;
      double aDist = Math.max(dist - MIN_DIST, 0);
      red = (int) (Math.max(0,
          255. * (aDist / ((SHORT_MAX_DIST + MIN_DIST) / 2.))));
    } else {
      red = 255;
      double aDist = Math.min((SHORT_MAX_DIST + MIN_DIST) / 2., dist
          - (SHORT_MAX_DIST + MIN_DIST) / 2.);
      green = (int) (Math.max(0,
          255 - 255. * (aDist / ((SHORT_MAX_DIST + MIN_DIST) / 2.))));
    }
    Log.v("GuideMe", "FlatArrow strokeColor: r=" + red + ",g=" + green + ",d="
        + dist);
    strokePaint.setARGB(0xC0, red, green, 0);
  }

  /**
   * Goes through the route step by step, whenever the distance to the current
   * point is less than MIN_DIST.
   */
  private void setNextRoutePoint() {
    if (route == null || route.length == 0) {
      // make sure there is a route
      next = -1;
      return;
    }
    // find the nearest route line
    double minDist = Double.POSITIVE_INFINITY;
    int minDistAt = 1;
    GeoCoordinate intersect = null;
    GeoCoordinate bestIntersect = null;
    for (int i = 0; i < route.length - 1; i++) {
      // find point on line that is the nearest to position
      intersect = getNearestIntersectWithLine(route[i], route[i + 1], location);
      double dist = location.distanceTo(intersect);
      if (dist < minDist) {
        // take the nearest one
        bestIntersect = intersect;
        minDist = dist;
        minDistAt = i + 1;
      }
    }
    next = minDistAt;
    // if we're at the end of the line, take a further step
    if (location.distanceTo(route[next]) * EARTH_RADIUS < MIN_DIST
        && next < route.length - 1)
      next++;

    // get the new track angle
    if (minDist < MIN_DIST)
      // for when we're on the wrong side of the street
      trackAngle = (float) (bestIntersect.trackAngle(route[next]) * GeoCoordinate.RAD_TO_DEG);
    else
      // for when we're far away from the route
      trackAngle = (float) (location.trackAngle(route[next]) * GeoCoordinate.RAD_TO_DEG);
  }

  /**
   * This method calculates the position on the straight line between to
   * locations that is the nearest to a specified location.
   * 
   * @param l1
   *          start point of line
   * @param l2
   *          end point of line
   * @param p
   *          location
   * @return nearest point to p on the segment between l1 and l2
   */
  private GeoCoordinate getNearestIntersectWithLine(GeoCoordinate l1,
      GeoCoordinate l2, GeoCoordinate p) {
    double E = 0.0001;
    double x1 = l1.getLongitude();
    double y1 = l1.getLatitude();
    double x2 = l2.getLongitude();
    double y2 = l2.getLatitude();
    double xp = p.getLongitude();
    double yp = p.getLatitude();
    GeoCoordinate result;
    if (Math.abs(x1 - x2) < E) {
      // easy case 1: same longitude
      double y = p.getLatitude();
      // stay inside segment
      if (y1 > y2) {
        if (y > y1)
          y = y1;
        if (y < y2)
          y = y2;
      }
      if (y1 < y2) {
        if (y < y1)
          y = y1;
        if (y > y2)
          y = y2;
      }
      result = new GeoCoordinate(y, x1);
    } else if (Math.abs(y1 - y2) < E) {
      // easy case 2: same longitude
      double x = p.getLongitude();
      // stay inside segment
      if (x1 > x2) {
        if (x > x1)
          x = x1;
        if (x < x2)
          x = x2;
      }
      if (x1 < x2) {
        if (x < x1)
          x = x1;
        if (x > x2)
          x = x2;
      }
      result = new GeoCoordinate(y1, x);
    } else {
      // the complicated case
      double ml = (y2 - y1) / (x2 - x1);
      double mp = -1. / ml;
      // intersect lines
      double xs = (mp * xp - ml * x1 + y1 - yp) / (mp - ml);
      double ys = mp * (xs - xp) + yp;
      // stay inside segment
      if (Math.abs(ml) <= 1.) {
        if (x1 > x2) {
          if (xs > x1) {
            xs = x1;
            ys = y1;
          }
          if (xs < x2) {
            xs = x2;
            ys = y2;
          }
        }
        if (x1 < x2) {
          if (xs < x1) {
            xs = x1;
            ys = y1;
          }
          if (xs > x2) {
            xs = x2;
            xs = x2;
          }
        }
      }
      if (Math.abs(ml) >= 1.) {
        if (y1 > y2) {
          if (ys > y1) {
            xs = x1;
            ys = y1;
          }
          if (ys < y2) {
            xs = x2;
            ys = y2;
          }
        }
        if (y1 < y2) {
          if (ys < y1) {
            xs = x1;
            ys = y1;
          }
          if (ys > y2) {
            xs = x2;
            xs = x2;
          }
        }
      }
      result = new GeoCoordinate(ys, xs);
    }

    return result;
  }
}
