package de.monarchy.guideme.view;

import static de.monarchy.guideme.util.GeoCoordinate.EARTH_RADIUS;
import static de.monarchy.guideme.util.GeoCoordinate.RAD_TO_DEG;

import java.util.HashMap;
import java.util.Iterator;
import java.util.Map;
import java.util.Set;

import android.app.Activity;
import android.view.View;
import android.view.ViewGroup;
import android.widget.AbsoluteLayout;
import android.widget.Toast;
import de.monarchy.guideme.util.GeoCoordinate;
import de.monarchy.guideme.util.POI;

/**
 * This class is a mixed view for camera preview and POIs. It uses an
 * AbsoluteLayout to position its members, which technically makes it
 * deprecated.
 * 
 * @author dstuere
 * 
 */
@SuppressWarnings("deprecation")
public class RealityView extends AbsoluteLayout {

  // The maximal angle a POI is allowed to differ from the view angle
  private static final int MAX_ANGLE = 45;

  // max distance to POI
  private static final int MAX_DIST = 5000;

  // POIs and their views
  private HashMap<POI, View> pois = new HashMap<POI, View>();
  // application context
  private Activity context;
  // view angle
  private int angle;
  // position
  private GeoCoordinate pos;
  // camera preview
  private CameraSurface cs;

  /**
   * Creates a new RealityView for the given application context.
   * 
   * @param context
   *          application context
   */
  public RealityView(Activity context) {
    // primary settings
    super(context);
    this.context = context;
    cs = CameraSurface.getInstance(context);
    angle = 0;
  }

  /**
   * Sets a new orientation. The orientation has to be in degrees, with 0 being
   * north, incrementing clockwise (east 90 degrees, south 180, west 270). The
   * orientation has to be delivered in device coordinates, the rotation of 90
   * degrees is measured inside the method.
   * 
   * @param arg
   *          new orientation
   */
  public void onOrientationChanged(float[] arg) {
    angle = ((int) arg[0] + 90) % 360;
    // move POIs
    relocate();
  }

  /**
   * Sets a new position. Checks for new POIs, deletes those in too great
   * distance and repositions the POIs on screen.
   * 
   * @param arg
   *          new location
   * @param arg1
   *          list of pois around the location
   */
  public void onPositionChanged(GeoCoordinate arg, POI[] arg1) {
    pos = arg;
    // remove bad POIs
    updatePOIMap();
    for (int i = 0; i < arg1.length; i++) {
      // add new POIs in distance less than
      if (!pois.containsKey(arg1[i])
          && arg1[i].getLocation().distanceTo(pos) * EARTH_RADIUS < MAX_DIST) {
        final POI a = arg1[i];
        View v = a.getView(context);
        pois.put(a, v);
        // if clicked, a short notification about the POI will appear
        v.setOnClickListener(new OnClickListener() {
          @Override
          public void onClick(View v) {
            Toast t = Toast.makeText(context, a.getDescription(),
                Toast.LENGTH_SHORT);
            t.show();
          }
        });
        // let the poi be available for routing
        context.registerForContextMenu(v);
      }
    }
    relocate();
  }

  /**
   * Deletes POIs which are too far away from the current position.
   */
  private void updatePOIMap() {
    Set<Map.Entry<POI, View>> es = pois.entrySet();
    Iterator<Map.Entry<POI, View>> iter = es.iterator();
    while (iter.hasNext()) {
      // check every entry for distance
      Map.Entry<POI, View> entry = iter.next();
      if (entry.getKey().getLocation().distanceTo(pos) * EARTH_RADIUS >= MAX_DIST) {
        // delete the POI view and the POI
        this.removeView(entry.getValue());
        context.unregisterForContextMenu(entry.getValue());
        iter.remove();
      }
    }
  }

  /**
   * Positions the POI views according to the view angle
   */
  private void relocate() {
    Set<Map.Entry<POI, View>> es = pois.entrySet();
    Iterator<Map.Entry<POI, View>> iter = es.iterator();
    int width = getWidth();
    while (iter.hasNext()) {
      // check every POI
      Map.Entry<POI, View> ent = iter.next();
      // calculate the angle between viewing angle and POI angle
      GeoCoordinate g = ent.getKey().getLocation();
      double trackAngle = pos.trackAngle(g) * RAD_TO_DEG;
      trackAngle = (trackAngle + 360.) % 360.;
      double angleFromView = trackAngle - angle;
      if (angleFromView < -180.)
        angleFromView += 360;

      if (Math.abs(angleFromView) > MAX_ANGLE) {
        // only show POIs within a certain angle
        this.removeView(ent.getValue());
      } else {
        // linearly move them on the screen
        int newX = (int) ((angleFromView / MAX_ANGLE) * width / 2);
        View v = ent.getValue();
        double distFactor = 3. - 2 * (pos.distanceTo(g) * EARTH_RADIUS / MAX_DIST);
        int viewWidth = (int) (16 * distFactor);
        int centre = width / 2 + newX - (viewWidth / 2);
        AbsoluteLayout.LayoutParams p = new AbsoluteLayout.LayoutParams(
            viewWidth, viewWidth, centre, (int)(viewWidth / 16. * viewWidth));
        if (v.getParent() == null) {
          // maybe they have to be added
          addView(v, p);
        } else {
          // or they just have to be moved
          v.setLayoutParams(p);
        }
        v.bringToFront();
      }
    }
    invalidate();
  }

  /**
   * Tidying up after closing the main activity.
   */
  public void onStop() {
    removeView(cs);
  }

  /**
   * Connecting the view at startup.
   */
  public void onStart() {
    // add camera to the layout
    this.addView(cs, 0, new AbsoluteLayout.LayoutParams(
        ViewGroup.LayoutParams.FILL_PARENT, ViewGroup.LayoutParams.FILL_PARENT,
        0, 0));
  }
}
